# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


import pickle
import timeit

import numpy as np

from pymatgen.core.composition import Composition
from pymatgen.core.lattice import Lattice
from pymatgen.core.periodic_table import Element, Species
from pymatgen.core.sites import PeriodicSite, Site
from pymatgen.electronic_structure.core import Magmom
from pymatgen.util.testing import PymatgenTest


class SiteTest(PymatgenTest):
    def setUp(self):
        self.ordered_site = Site("Fe", [0.25, 0.35, 0.45])
        self.disordered_site = Site({"Fe": 0.5, "Mn": 0.5}, [0.25, 0.35, 0.45])
        self.propertied_site = Site("Fe2+", [0.25, 0.35, 0.45], {"magmom": 5.1, "charge": 4.2})
        self.propertied_magmomvector_site = Site(
            "Fe2+",
            [0.25, 0.35, 0.45],
            {"magmom": Magmom([2.6, 2.6, 3.5]), "charge": 4.2},
        )
        self.dummy_site = Site("X", [0, 0, 0])

    def test_properties(self):
        self.assertRaises(AttributeError, getattr, self.disordered_site, "specie")
        self.assertIsInstance(self.ordered_site.specie, Element)
        self.assertEqual(self.propertied_site.properties["magmom"], 5.1)
        self.assertEqual(self.propertied_site.properties["charge"], 4.2)

    def test_to_from_dict(self):
        d = self.disordered_site.as_dict()
        site = Site.from_dict(d)
        self.assertEqual(site, self.disordered_site)
        self.assertNotEqual(site, self.ordered_site)
        d = self.propertied_site.as_dict()
        site = Site.from_dict(d)
        self.assertEqual(site.properties["magmom"], 5.1)
        self.assertEqual(site.properties["charge"], 4.2)
        d = self.propertied_magmomvector_site.as_dict()
        site = Site.from_dict(d)
        self.assertEqual(site.properties["magmom"], Magmom([2.6, 2.6, 3.5]))
        self.assertEqual(site.properties["charge"], 4.2)
        d = self.dummy_site.as_dict()
        site = Site.from_dict(d)
        self.assertEqual(site.species, self.dummy_site.species)

    def test_hash(self):
        self.assertEqual(self.ordered_site.__hash__(), 26)
        self.assertEqual(self.disordered_site.__hash__(), 51)

    def test_cmp(self):
        self.assertTrue(self.ordered_site > self.disordered_site)

    def test_distance(self):
        osite = self.ordered_site
        self.assertAlmostEqual(np.linalg.norm([0.25, 0.35, 0.45]), osite.distance_from_point([0, 0, 0]))
        self.assertAlmostEqual(osite.distance(self.disordered_site), 0)

    def test_pickle(self):
        o = pickle.dumps(self.propertied_site)
        self.assertEqual(pickle.loads(o), self.propertied_site)

    def test_setters(self):
        self.disordered_site.species = "Cu"
        self.assertEqual(self.disordered_site.species, Composition("Cu"))
        self.disordered_site.x = 1.25
        self.disordered_site.y = 1.35
        self.assertEqual(self.disordered_site.coords[0], 1.25)
        self.assertEqual(self.disordered_site.coords[1], 1.35)

        def set_bad_species():
            self.disordered_site.species = {"Cu": 0.5, "Gd": 0.6}

        self.assertRaises(ValueError, set_bad_species)


class PeriodicSiteTest(PymatgenTest):
    def setUp(self):
        self.lattice = Lattice.cubic(10.0)
        self.si = Element("Si")
        self.site = PeriodicSite("Fe", [0.25, 0.35, 0.45], self.lattice)
        self.site2 = PeriodicSite({"Si": 0.5}, [0, 0, 0], self.lattice)
        self.assertEqual(
            self.site2.species,
            Composition({Element("Si"): 0.5}),
            "Inconsistent site created!",
        )
        self.propertied_site = PeriodicSite(
            Species("Fe", 2),
            [0.25, 0.35, 0.45],
            self.lattice,
            properties={"magmom": 5.1, "charge": 4.2},
        )
        self.dummy_site = PeriodicSite("X", [0, 0, 0], self.lattice)

    def test_properties(self):
        """
        Test the properties for a site
        """
        self.assertEqual(self.site.a, 0.25)
        self.assertEqual(self.site.b, 0.35)
        self.assertEqual(self.site.c, 0.45)
        self.assertEqual(self.site.x, 2.5)
        self.assertEqual(self.site.y, 3.5)
        self.assertEqual(self.site.z, 4.5)
        self.assertTrue(self.site.is_ordered)
        self.assertFalse(self.site2.is_ordered)
        self.assertEqual(self.propertied_site.properties["magmom"], 5.1)
        self.assertEqual(self.propertied_site.properties["charge"], 4.2)

    def test_distance(self):
        other_site = PeriodicSite("Fe", np.array([0, 0, 0]), self.lattice)
        self.assertAlmostEqual(self.site.distance(other_site), 6.22494979899, 5)

    def test_distance_from_point(self):
        self.assertNotAlmostEqual(self.site.distance_from_point([0.1, 0.1, 0.1]), 6.22494979899, 5)
        self.assertAlmostEqual(self.site.distance_from_point([0.1, 0.1, 0.1]), 6.0564015718906887, 5)

    def test_distance_and_image(self):
        other_site = PeriodicSite("Fe", np.array([1, 1, 1]), self.lattice)
        (distance, image) = self.site.distance_and_image(other_site)
        self.assertAlmostEqual(distance, 6.22494979899, 5)
        self.assertTrue(([-1, -1, -1] == image).all())
        (distance, image) = self.site.distance_and_image(other_site, [1, 0, 0])
        self.assertAlmostEqual(distance, 19.461500456028563, 5)
        # Test that old and new distance algo give the same ans for
        # "standard lattices"
        lattice = Lattice(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]))
        site1 = PeriodicSite("Fe", np.array([0.01, 0.02, 0.03]), lattice)
        site2 = PeriodicSite("Fe", np.array([0.99, 0.98, 0.97]), lattice)
        self.assertAlmostEqual(
            get_distance_and_image_old(site1, site2)[0],
            site1.distance_and_image(site2)[0],
        )
        lattice = Lattice.from_parameters(1, 0.01, 1, 10, 10, 10)
        site1 = PeriodicSite("Fe", np.array([0.01, 0.02, 0.03]), lattice)
        site2 = PeriodicSite("Fe", np.array([0.99, 0.98, 0.97]), lattice)
        self.assertTrue(get_distance_and_image_old(site1, site2)[0] > site1.distance_and_image(site2)[0])
        site2 = PeriodicSite("Fe", np.random.rand(3), lattice)
        (dist_old, jimage_old) = get_distance_and_image_old(site1, site2)
        (dist_new, jimage_new) = site1.distance_and_image(site2)
        self.assertTrue(
            dist_old - dist_new > -1e-8,
            "New distance algo should give smaller answers!",
        )
        self.assertFalse(
            (abs(dist_old - dist_new) < 1e-8) ^ (jimage_old == jimage_new).all(),
            "If old dist == new dist, images must be the same!",
        )
        latt = Lattice.from_parameters(3.0, 3.1, 10.0, 2.96, 2.0, 1.0)
        site = PeriodicSite("Fe", [0.1, 0.1, 0.1], latt)
        site2 = PeriodicSite("Fe", [0.99, 0.99, 0.99], latt)
        (dist, img) = site.distance_and_image(site2)
        self.assertAlmostEqual(dist, 0.15495358379511573)
        self.assertEqual(list(img), [-11, 6, 0])

    def test_is_periodic_image(self):
        other = PeriodicSite("Fe", np.array([1.25, 2.35, 4.45]), self.lattice)
        self.assertTrue(
            self.site.is_periodic_image(other),
            "This other site should be a periodic image.",
        )
        other = PeriodicSite("Fe", np.array([1.25, 2.35, 4.46]), self.lattice)
        self.assertFalse(
            self.site.is_periodic_image(other),
            "This other site should not be a periodic image.",
        )
        other = PeriodicSite("Fe", np.array([1.25, 2.35, 4.45]), Lattice.rhombohedral(2, 60))
        self.assertFalse(
            self.site.is_periodic_image(other),
            "Different lattices should not be periodic images.",
        )

    def test_equality(self):
        other_site = PeriodicSite("Fe", np.array([1, 1, 1]), self.lattice)
        self.assertTrue(self.site.__eq__(self.site))
        self.assertFalse(other_site.__eq__(self.site))
        self.assertFalse(self.site.__ne__(self.site))
        self.assertTrue(other_site.__ne__(self.site))

    def test_as_from_dict(self):
        d = self.site2.as_dict()
        site = PeriodicSite.from_dict(d)
        self.assertEqual(site, self.site2)
        self.assertNotEqual(site, self.site)
        d = self.propertied_site.as_dict()
        site3 = PeriodicSite({"Si": 0.5, "Fe": 0.5}, [0, 0, 0], self.lattice)
        d = site3.as_dict()
        site = PeriodicSite.from_dict(d)
        self.assertEqual(site.species, site3.species)

        d = self.dummy_site.as_dict()
        site = PeriodicSite.from_dict(d)
        self.assertEqual(site.species, self.dummy_site.species)

    def test_to_unit_cell(self):
        site = PeriodicSite("Fe", np.array([1.25, 2.35, 4.46]), self.lattice)
        site.to_unit_cell(in_place=True)
        val = [0.25, 0.35, 0.46]
        self.assertArrayAlmostEqual(site.frac_coords, val)

    def test_setters(self):
        site = self.propertied_site
        site.species = "Cu"
        self.assertEqual(site.species, Composition("Cu"))
        site.x = 1.25
        site.y = 1.35
        self.assertEqual(site.coords[0], 1.25)
        self.assertEqual(site.coords[1], 1.35)
        self.assertEqual(site.a, 0.125)
        self.assertEqual(site.b, 0.135)
        site.lattice = Lattice.cubic(100)
        self.assertEqual(site.x, 12.5)

        def set_bad_species():
            site.species = {"Cu": 0.5, "Gd": 0.6}

        self.assertRaises(ValueError, set_bad_species)

        site.frac_coords = [0, 0, 0.1]
        self.assertArrayAlmostEqual(site.coords, [0, 0, 10])
        site.coords = [1.5, 3.25, 5]
        self.assertArrayAlmostEqual(site.frac_coords, [0.015, 0.0325, 0.05])

    def test_repr(self):
        self.assertEqual(
            self.propertied_site.__repr__(),
            "PeriodicSite: Fe2+ (2.5000, 3.5000, 4.5000) [0.2500, 0.3500, 0.4500]",
        )


def get_distance_and_image_old(site1, site2, jimage=None):
    """
    Gets distance between two sites assuming periodic boundary conditions.
    If the index jimage of two sites atom j is not specified it selects the
    j image nearest to the i atom and returns the distance and jimage
    indices in terms of lattice vector translations. If the index jimage of
    atom j is specified it returns the distance between the i atom and the
    specified jimage atom, the given jimage is also returned.

    Args:
        other:
            other site to get distance from.
        jimage:
            specific periodic image in terms of lattice translations,
            e.g., [1,0,0] implies to take periodic image that is one
            a-lattice vector away. If jimage is None, the image that is
            nearest to the site is found.

    Returns:
        (distance, jimage):
            distance and periodic lattice translations of the other site
            for which the distance applies.

    .. note::
        Assumes the primitive cell vectors are sufficiently not skewed such
        that the condition \\|a\\|cos(ab_angle) < \\|b\\| for all possible cell
        vector pairs. ** this method does not check this condition **
    """
    if jimage is None:
        # Old algorithm
        jimage = -np.array(np.around(site2.frac_coords - site1.frac_coords), int)
    mapped_vec = site1.lattice.get_cartesian_coords(jimage + site2.frac_coords - site1.frac_coords)
    dist = np.linalg.norm(mapped_vec)
    return dist, jimage


if __name__ == "__main__":
    import unittest

    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.
"""
Created on Nov 10, 2012

@author: shyue
"""

from pymatgen.util.testing import PymatgenTest

import random
import unittest

from pymatgen.core.composition import ChemicalPotential, Composition
from pymatgen.core.periodic_table import Element, Species


class CompositionTest(PymatgenTest):
    def setUp(self):
        self.comp = [
            Composition("Li3Fe2(PO4)3"),
            Composition("Li3Fe(PO4)O"),
            Composition("LiMn2O4"),
            Composition("Li4O4"),
            Composition("Li3Fe2Mo3O12"),
            Composition("Li3Fe2((PO4)3(CO3)5)2"),
            Composition("Li1.5Si0.5"),
            Composition("ZnOH"),
        ]

        self.indeterminate_comp = []
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("Co1", True))
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("Co1", False))
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("co2o3"))
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("ncalu"))
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("calun"))
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("liCoo2n (pO4)2"))
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("(co)2 (PO)4"))
        self.indeterminate_comp.append(Composition.ranked_compositions_from_indeterminate_formula("Fee3"))

    def test_immutable(self):
        try:
            self.comp[0]["Fe"] = 1
        except Exception as ex:
            self.assertIsInstance(ex, TypeError)

        try:
            del self.comp[0]["Fe"]
        except Exception as ex:
            self.assertIsInstance(ex, TypeError)

    def test_in(self):
        self.assertIn("Fe", self.comp[0])
        self.assertNotIn("Fe", self.comp[2])
        self.assertIn(Element("Fe"), self.comp[0])
        self.assertEqual(self.comp[0]["Fe"], 2)
        self.assertEqual(self.comp[0]["Mn"], 0)
        self.assertRaises(TypeError, self.comp[0].__getitem__, "Hello")
        self.assertRaises(TypeError, self.comp[0].__getitem__, "Vac")

    def test_hill_formula(self):
        c = Composition("CaCO3")
        self.assertEqual(c.hill_formula, "C Ca O3")
        c = Composition("C2H5OH")
        self.assertEqual(c.hill_formula, "C2 H6 O")

    def test_init_(self):
        self.assertRaises(ValueError, Composition, {"H": -0.1})
        f = {"Fe": 4, "Li": 4, "O": 16, "P": 4}
        self.assertEqual("Li4 Fe4 P4 O16", Composition(f).formula)
        f = {None: 4, "Li": 4, "O": 16, "P": 4}
        self.assertRaises(TypeError, Composition, f)
        f = {1: 2, 8: 1}
        self.assertEqual("H2 O1", Composition(f).formula)
        self.assertEqual("Na2 O1", Composition(Na=2, O=1).formula)

        c = Composition({"S": Composition.amount_tolerance / 2})
        self.assertEqual(len(c.elements), 0)

    def test_average_electroneg(self):
        val = [
            2.7224999999999997,
            2.4160000000000004,
            2.5485714285714285,
            2.21,
            2.718,
            3.08,
            1.21,
            2.43,
        ]
        for i, c in enumerate(self.comp):
            self.assertAlmostEqual(c.average_electroneg, val[i])

    def test_total_electrons(self):
        test_cases = {"C": 6, "SrTiO3": 84}
        for item in test_cases.keys():
            c = Composition(item)
            self.assertAlmostEqual(c.total_electrons, test_cases[item])

    def test_formula(self):
        correct_formulas = [
            "Li3 Fe2 P3 O12",
            "Li3 Fe1 P1 O5",
            "Li1 Mn2 O4",
            "Li4 O4",
            "Li3 Fe2 Mo3 O12",
            "Li3 Fe2 P6 C10 O54",
            "Li1.5 Si0.5",
            "Zn1 H1 O1",
        ]
        all_formulas = [c.formula for c in self.comp]
        self.assertEqual(all_formulas, correct_formulas)
        self.assertRaises(ValueError, Composition, "(co2)(po4)2")

        self.assertEqual(Composition("K Na 2").reduced_formula, "KNa2")

        self.assertEqual(Composition("K3 Na 2").reduced_formula, "K3Na2")

        self.assertEqual(Composition("Na 3 Zr (PO 4) 3").reduced_formula, "Na3Zr(PO4)3")

    def test_to_latex_html_unicode(self):
        self.assertEqual(self.comp[0].to_latex_string(), "Li$_{3}$Fe$_{2}$P$_{3}$O$_{12}$")
        self.assertEqual(self.comp[0].to_html_string(), "Li<sub>3</sub>Fe<sub>2</sub>P<sub>3</sub>O<sub>12</sub>")
        self.assertEqual(self.comp[0].to_unicode_string(), "Li₃Fe₂P₃O₁₂")

    def test_iupac_formula(self):
        correct_formulas = [
            "Li3 Fe2 P3 O12",
            "Li3 Fe1 P1 O5",
            "Li1 Mn2 O4",
            "Li4 O4",
            "Li3 Mo3 Fe2 O12",
            "Li3 Fe2 C10 P6 O54",
            "Li1.5 Si0.5",
            "Zn1 H1 O1",
        ]
        all_formulas = [c.iupac_formula for c in self.comp]
        self.assertEqual(all_formulas, correct_formulas)

    def test_mixed_valence(self):
        comp = Composition({"Fe2+": 2, "Fe3+": 4, "Li+": 8})
        self.assertEqual(comp.reduced_formula, "Li4Fe3")
        self.assertEqual(comp.alphabetical_formula, "Fe6 Li8")
        self.assertEqual(comp.formula, "Li8 Fe6")

    def test_indeterminate_formula(self):
        correct_formulas = [
            ["Co1"],
            ["Co1", "C1 O1"],
            ["Co2 O3", "C1 O5"],
            ["N1 Ca1 Lu1", "U1 Al1 C1 N1"],
            ["N1 Ca1 Lu1", "U1 Al1 C1 N1"],
            [
                "Li1 Co1 P2 N1 O10",
                "Li1 Co1 Po8 N1 O2",
                "Li1 P2 C1 N1 O11",
                "Li1 Po8 C1 N1 O3",
            ],
            ["Co2 P4 O4", "Co2 Po4", "P4 C2 O6", "Po4 C2 O2"],
            [],
        ]
        for i, c in enumerate(correct_formulas):
            self.assertEqual([Composition(comp) for comp in c], self.indeterminate_comp[i])

    def test_alphabetical_formula(self):
        correct_formulas = [
            "Fe2 Li3 O12 P3",
            "Fe1 Li3 O5 P1",
            "Li1 Mn2 O4",
            "Li4 O4",
            "Fe2 Li3 Mo3 O12",
            "C10 Fe2 Li3 O54 P6",
            "Li1.5 Si0.5",
            "H1 O1 Zn1",
        ]
        all_formulas = [c.alphabetical_formula for c in self.comp]
        self.assertEqual(all_formulas, correct_formulas)

    def test_reduced_composition(self):
        correct_reduced_formulas = [
            "Li3Fe2(PO4)3",
            "Li3FePO5",
            "LiMn2O4",
            "Li2O2",
            "Li3Fe2(MoO4)3",
            "Li3Fe2P6(C5O27)2",
            "Li1.5Si0.5",
            "ZnHO",
        ]
        for i in range(len(self.comp)):
            self.assertEqual(
                self.comp[i].get_reduced_composition_and_factor()[0],
                Composition(correct_reduced_formulas[i]),
            )

    def test_reduced_formula(self):
        correct_reduced_formulas = [
            "Li3Fe2(PO4)3",
            "Li3FePO5",
            "LiMn2O4",
            "Li2O2",
            "Li3Fe2(MoO4)3",
            "Li3Fe2P6(C5O27)2",
            "Li1.5Si0.5",
            "ZnHO",
        ]
        all_formulas = [c.reduced_formula for c in self.comp]
        self.assertEqual(all_formulas, correct_reduced_formulas)

        # test iupac reduced formula (polyanions should still appear at the end)
        all_formulas = [c.get_reduced_formula_and_factor(iupac_ordering=True)[0] for c in self.comp]
        self.assertEqual(all_formulas, correct_reduced_formulas)
        self.assertEqual(
            Composition("H6CN").get_integer_formula_and_factor(iupac_ordering=True)[0],
            "CNH6",
        )

        # test rounding
        c = Composition({"Na": 2 - Composition.amount_tolerance / 2, "Cl": 2})
        self.assertEqual("NaCl", c.reduced_formula)

    def test_integer_formula(self):
        correct_reduced_formulas = [
            "Li3Fe2(PO4)3",
            "Li3FePO5",
            "LiMn2O4",
            "Li2O2",
            "Li3Fe2(MoO4)3",
            "Li3Fe2P6(C5O27)2",
            "Li3Si",
            "ZnHO",
        ]
        all_formulas = [c.get_integer_formula_and_factor()[0] for c in self.comp]
        self.assertEqual(all_formulas, correct_reduced_formulas)
        self.assertEqual(Composition("Li0.5O0.25").get_integer_formula_and_factor(), ("Li2O", 0.25))
        self.assertEqual(Composition("O0.25").get_integer_formula_and_factor(), ("O2", 0.125))
        formula, factor = Composition("Li0.16666667B1.0H1.0").get_integer_formula_and_factor()
        self.assertEqual(formula, "Li(BH)6")
        self.assertAlmostEqual(factor, 1 / 6)

        # test iupac reduced formula (polyanions should still appear at the end)
        all_formulas = [c.get_integer_formula_and_factor(iupac_ordering=True)[0] for c in self.comp]
        self.assertEqual(all_formulas, correct_reduced_formulas)
        self.assertEqual(
            Composition("H6CN0.5").get_integer_formula_and_factor(iupac_ordering=True),
            ("C2NH12", 0.5),
        )

    def test_num_atoms(self):
        correct_num_atoms = [20, 10, 7, 8, 20, 75, 2, 3]

        all_natoms = [c.num_atoms for c in self.comp]
        self.assertEqual(all_natoms, correct_num_atoms)

    def test_weight(self):
        correct_weights = [
            417.427086,
            187.63876199999999,
            180.81469,
            91.7616,
            612.3258,
            1302.430172,
            24.454250000000002,
            82.41634,
        ]
        all_weights = [c.weight for c in self.comp]
        self.assertArrayAlmostEqual(all_weights, correct_weights, 5)

    def test_get_atomic_fraction(self):
        correct_at_frac = {"Li": 0.15, "Fe": 0.1, "P": 0.15, "O": 0.6}
        for el in ["Li", "Fe", "P", "O"]:
            self.assertEqual(
                self.comp[0].get_atomic_fraction(el),
                correct_at_frac[el],
                "Wrong computed atomic fractions",
            )
        self.assertEqual(self.comp[0].get_atomic_fraction("S"), 0, "Wrong computed atomic fractions")

    def test_anonymized_formula(self):
        expected_formulas = [
            "A2B3C3D12",
            "ABC3D5",
            "AB2C4",
            "AB",
            "A2B3C3D12",
            "A2B3C6D10E54",
            "A0.5B1.5",
            "ABC",
        ]
        for i in range(len(self.comp)):
            self.assertEqual(self.comp[i].anonymized_formula, expected_formulas[i])

    def test_get_wt_fraction(self):
        correct_wt_frac = {
            "Li": 0.0498841610868,
            "Fe": 0.267567687258,
            "P": 0.222604831158,
            "O": 0.459943320496,
        }
        for el in ["Li", "Fe", "P", "O"]:
            self.assertAlmostEqual(
                correct_wt_frac[el],
                self.comp[0].get_wt_fraction(el),
                5,
                "Wrong computed weight fraction",
            )
        self.assertEqual(
            self.comp[0].get_wt_fraction(Element("S")),
            0,
            "Wrong computed weight fractions",
        )

    def test_from_dict(self):
        sym_dict = {"Fe": 6, "O": 8}
        self.assertEqual(
            Composition.from_dict(sym_dict).reduced_formula,
            "Fe3O4",
            "Creation form sym_amount dictionary failed!",
        )
        comp = Composition({"Fe2+": 2, "Fe3+": 4, "O2-": 8})
        comp2 = Composition.from_dict(comp.as_dict())
        self.assertEqual(comp, comp2)

    def test_as_dict(self):
        c = Composition.from_dict({"Fe": 4, "O": 6})
        d = c.as_dict()
        correct_dict = {"Fe": 4.0, "O": 6.0}
        self.assertEqual(d["Fe"], correct_dict["Fe"])
        self.assertEqual(d["O"], correct_dict["O"])
        correct_dict = {"Fe": 2.0, "O": 3.0}
        d = c.to_reduced_dict
        self.assertIsInstance(d, dict)
        self.assertEqual(d["Fe"], correct_dict["Fe"])
        self.assertEqual(d["O"], correct_dict["O"])

    def test_pickle(self):
        for c in self.comp:
            self.serialize_with_pickle(c, test_eq=True)
            self.serialize_with_pickle(c.to_data_dict, test_eq=True)

    def test_to_data_dict(self):
        comp = Composition("Fe0.00009Ni0.99991")
        d = comp.to_data_dict
        self.assertAlmostEqual(d["reduced_cell_composition"]["Fe"], 9e-5)

    def test_add(self):
        self.assertEqual(
            (self.comp[0] + self.comp[2]).formula,
            "Li4 Mn2 Fe2 P3 O16",
            "Incorrect composition after addition!",
        )
        self.assertEqual(
            (self.comp[3] + {"Fe": 4, "O": 4}).formula,
            "Li4 Fe4 O8",
            "Incorrect composition after addition!",
        )

    def test_sub(self):
        self.assertEqual(
            (self.comp[0] - Composition("Li2O")).formula,
            "Li1 Fe2 P3 O11",
            "Incorrect composition after addition!",
        )
        self.assertEqual((self.comp[0] - {"Fe": 2, "O": 3}).formula, "Li3 P3 O9")

        self.assertRaises(ValueError, Composition("O").__sub__, Composition("H"))

        # check that S is completely removed by subtraction
        c1 = Composition({"S": 1 + Composition.amount_tolerance / 2, "O": 1})
        c2 = Composition({"S": 1})
        self.assertEqual(len((c1 - c2).elements), 1)

    def test_mul(self):
        self.assertEqual((self.comp[0] * 4).formula, "Li12 Fe8 P12 O48")
        self.assertEqual((3 * self.comp[1]).formula, "Li9 Fe3 P3 O15")

    def test_div(self):
        self.assertEqual((self.comp[0] / 4).formula, "Li0.75 Fe0.5 P0.75 O3")

    def test_equals(self):
        random_z = random.randint(1, 92)
        fixed_el = Element.from_Z(random_z)
        other_z = random.randint(1, 92)
        while other_z == random_z:
            other_z = random.randint(1, 92)
        comp1 = Composition({fixed_el: 1, Element.from_Z(other_z): 0})
        other_z = random.randint(1, 92)
        while other_z == random_z:
            other_z = random.randint(1, 92)
        comp2 = Composition({fixed_el: 1, Element.from_Z(other_z): 0})
        self.assertEqual(
            comp1,
            comp2,
            "Composition equality test failed. " + "%s should be equal to %s" % (comp1.formula, comp2.formula),
        )
        self.assertEqual(comp1.__hash__(), comp2.__hash__(), "Hashcode equality test failed!")

    def test_hash_robustness(self):
        c1 = Composition(f"O{0.2}Fe{0.8}Na{Composition.amount_tolerance*0.99}")
        c2 = Composition(f"O{0.2}Fe{0.8}Na{Composition.amount_tolerance*1.01}")
        c3 = Composition(f"O{0.2}Fe{0.8+Composition.amount_tolerance*0.99}")

        self.assertTrue(c1 == c3, "__eq__ not robust")
        self.assertEqual(c1 == c3, hash(c1) == hash(c3), "Hash doesn't match eq when true")
        self.assertFalse(hash(c1) == hash(c2), "Hash equal for different chemical systems")

    def test_comparisons(self):
        c1 = Composition({"S": 1})
        c1_1 = Composition({"S": 1.00000000000001})
        c2 = Composition({"S": 2})
        c3 = Composition({"O": 1})
        c4 = Composition({"O": 1, "S": 1})
        self.assertFalse(c1 > c2)
        self.assertFalse(c1_1 > c1)
        self.assertFalse(c1_1 < c1)
        self.assertTrue(c1 > c3)
        self.assertTrue(c3 < c1)
        self.assertTrue(c4 > c1)
        self.assertEqual(sorted([c1, c1_1, c2, c4, c3]), [c3, c1, c1_1, c4, c2])

    def test_almost_equals(self):
        c1 = Composition({"Fe": 2.0, "O": 3.0, "Mn": 0})
        c2 = Composition({"O": 3.2, "Fe": 1.9, "Zn": 0})
        c3 = Composition({"Ag": 2.0, "O": 3.0})
        c4 = Composition({"Fe": 2.0, "O": 3.0, "Ag": 2.0})
        self.assertTrue(c1.almost_equals(c2, rtol=0.1))
        self.assertFalse(c1.almost_equals(c2, rtol=0.01))
        self.assertFalse(c1.almost_equals(c3, rtol=0.1))
        self.assertFalse(c1.almost_equals(c4, rtol=0.1))

    def test_equality(self):
        self.assertTrue(self.comp[0].__eq__(self.comp[0]))
        self.assertFalse(self.comp[0].__eq__(self.comp[1]))
        self.assertFalse(self.comp[0].__ne__(self.comp[0]))
        self.assertTrue(self.comp[0].__ne__(self.comp[1]))

    def test_fractional_composition(self):
        for c in self.comp:
            self.assertAlmostEqual(c.fractional_composition.num_atoms, 1)

    def test_init_numerical_tolerance(self):
        self.assertEqual(Composition({"B": 1, "C": -1e-12}), Composition("B"))

    def test_negative_compositions(self):
        self.assertEqual(Composition("Li-1(PO-1)4", allow_negative=True).formula, "Li-1 P4 O-4")
        self.assertEqual(
            Composition("Li-1(PO-1)4", allow_negative=True).reduced_formula,
            "Li-1(PO-1)4",
        )
        self.assertEqual(
            Composition("Li-2Mg4", allow_negative=True).reduced_composition,
            Composition("Li-1Mg2", allow_negative=True),
        )
        self.assertEqual(
            Composition("Li-2.5Mg4", allow_negative=True).reduced_composition,
            Composition("Li-2.5Mg4", allow_negative=True),
        )

        # test math
        c1 = Composition("LiCl", allow_negative=True)
        c2 = Composition("Li")
        self.assertEqual(c1 - 2 * c2, Composition({"Li": -1, "Cl": 1}, allow_negative=True))
        self.assertEqual((c1 + c2).allow_negative, True)
        self.assertEqual(c1 / -1, Composition("Li-1Cl-1", allow_negative=True))

        # test num_atoms
        c1 = Composition("Mg-1Li", allow_negative=True)
        self.assertEqual(c1.num_atoms, 2)
        self.assertEqual(c1.get_atomic_fraction("Mg"), 0.5)
        self.assertEqual(c1.get_atomic_fraction("Li"), 0.5)
        self.assertEqual(c1.fractional_composition, Composition("Mg-0.5Li0.5", allow_negative=True))

        # test copy
        self.assertEqual(c1.copy(), c1)

        # test species
        c1 = Composition({"Mg": 1, "Mg2+": -1}, allow_negative=True)
        self.assertEqual(c1.num_atoms, 2)
        self.assertEqual(c1.element_composition, Composition())
        self.assertEqual(c1.average_electroneg, 1.31)

    def test_special_formulas(self):
        special_formulas = {
            "LiO": "Li2O2",
            "NaO": "Na2O2",
            "KO": "K2O2",
            "HO": "H2O2",
            "CsO": "Cs2O2",
            "RbO": "Rb2O2",
            "O": "O2",
            "N": "N2",
            "F": "F2",
            "Cl": "Cl2",
            "H": "H2",
        }
        for k, v in special_formulas.items():
            self.assertEqual(Composition(k).reduced_formula, v)

    def test_oxi_state_guesses(self):
        self.assertEqual(Composition("LiFeO2").oxi_state_guesses(), ({"Li": 1, "Fe": 3, "O": -2},))

        self.assertEqual(Composition("Fe4O5").oxi_state_guesses(), ({"Fe": 2.5, "O": -2},))

        self.assertEqual(Composition("V2O3").oxi_state_guesses(), ({"V": 3, "O": -2},))

        # all_oxidation_states produces *many* possible responses
        self.assertEqual(len(Composition("MnO").oxi_state_guesses(all_oxi_states=True)), 4)

        # can't balance b/c missing V4+
        self.assertEqual(
            Composition("VO2").oxi_state_guesses(oxi_states_override={"V": [2, 3, 5]}),
            [],
        )

        # missing V4+, but can balance due to additional sites
        self.assertEqual(
            Composition("V2O4").oxi_state_guesses(oxi_states_override={"V": [2, 3, 5]}),
            ({"V": 4, "O": -2},),
        )

        # multiple solutions - Mn/Fe = 2+/4+ or 3+/3+ or 4+/2+
        self.assertEqual(
            len(Composition("MnFeO3").oxi_state_guesses(oxi_states_override={"Mn": [2, 3, 4], "Fe": [2, 3, 4]})),
            3,
        )

        # multiple solutions prefers 3/3 over 2/4 or 4/2
        self.assertEqual(
            Composition("MnFeO3").oxi_state_guesses(oxi_states_override={"Mn": [2, 3, 4], "Fe": [2, 3, 4]})[0],
            {"Mn": 3, "Fe": 3, "O": -2},
        )

        # target charge of 1
        self.assertEqual(
            Composition("V2O6").oxi_state_guesses(oxi_states_override={"V": [2, 3, 4, 5]}, target_charge=-2),
            ({"V": 5, "O": -2},),
        )

        # max_sites for very large composition - should timeout if incorrect
        self.assertEqual(
            Composition("Li10000Fe10000P10000O40000").oxi_state_guesses(max_sites=7)[0],
            {"Li": 1, "Fe": 2, "P": 5, "O": -2},
        )

        # max_sites for very large composition - should timeout if incorrect
        self.assertEqual(
            Composition("Li10000Fe10000P10000O40000").oxi_state_guesses(max_sites=-1)[0],
            {"Li": 1, "Fe": 2, "P": 5, "O": -2},
        )

        # negative max_sites less than -1 - should throw error if cannot reduce
        # to under the abs(max_sites) number of sites. Will also timeout if
        # incorrect.
        self.assertEqual(
            Composition("Sb10000O10000F10000").oxi_state_guesses(max_sites=-3)[0],
            {"Sb": 3, "O": -2, "F": -1},
        )
        self.assertRaises(ValueError, Composition("LiOF").oxi_state_guesses, max_sites=-2)

        self.assertRaises(ValueError, Composition("V2O3").oxi_state_guesses, max_sites=1)

    def test_oxi_state_decoration(self):
        # Basic test: Get compositions where each element is in a single charge state
        decorated = Composition("H2O").add_charges_from_oxi_state_guesses()
        self.assertIn(Species("H", 1), decorated)
        self.assertEqual(2, decorated.get(Species("H", 1)))

        # Test: More than one charge state per element
        decorated = Composition("Fe3O4").add_charges_from_oxi_state_guesses()
        self.assertEqual(1, decorated.get(Species("Fe", 2)))
        self.assertEqual(2, decorated.get(Species("Fe", 3)))
        self.assertEqual(4, decorated.get(Species("O", -2)))

        # Test: No possible charge states
        #   It should return an uncharged composition
        decorated = Composition("NiAl").add_charges_from_oxi_state_guesses()
        self.assertEqual(1, decorated.get(Species("Ni", 0)))
        self.assertEqual(1, decorated.get(Species("Al", 0)))

    def test_Metallofullerene(self):
        # Test: Parse Metallofullerene formula (e.g. Y3N@C80)
        formula = "Y3N@C80"
        sym_dict = {"Y": 3, "N": 1, "C": 80}
        cmp = Composition(formula)
        cmp2 = Composition.from_dict(sym_dict)
        self.assertEqual(cmp, cmp2)

    def test_contains_element_type(self):

        formula = "EuTiO3"
        cmp = Composition(formula)
        self.assertTrue(cmp.contains_element_type("lanthanoid"))
        self.assertFalse(cmp.contains_element_type("noble_gas"))
        self.assertTrue(cmp.contains_element_type("f-block"))
        self.assertFalse(cmp.contains_element_type("s-block"))

    def test_chemical_system(self):

        formula = "NaCl"
        cmp = Composition(formula)
        self.assertEqual(cmp.chemical_system, "Cl-Na")

    def test_is_valid(self):

        formula = "NaCl"
        cmp = Composition(formula)
        self.assertTrue(cmp.valid)

        formula = "NaClX"
        cmp = Composition(formula)
        self.assertFalse(cmp.valid)

        self.assertRaises(ValueError, Composition, "NaClX", strict=True)

    def test_remove_charges(self):
        cmp1 = Composition({"Al3+": 2.0, "O2-": 3.0})

        cmp2 = Composition({"Al": 2.0, "O": 3.0})
        self.assertNotEqual(str(cmp1), str(cmp2))

        cmp1 = cmp1.remove_charges()
        self.assertEqual(str(cmp1), str(cmp2))

        cmp1 = cmp1.remove_charges()
        self.assertEqual(str(cmp1), str(cmp2))

        cmp1 = Composition({"Fe3+": 2.0, "Fe2+": 3.0, "O2-": 6.0})
        cmp2 = Composition({"Fe": 5.0, "O": 6.0})
        self.assertNotEqual(str(cmp1), str(cmp2))

        cmp1 = cmp1.remove_charges()
        self.assertEqual(str(cmp1), str(cmp2))


class ChemicalPotentialTest(unittest.TestCase):
    def test_init(self):
        d = {"Fe": 1, Element("Fe"): 1}
        self.assertRaises(ValueError, ChemicalPotential, d)
        for k in ChemicalPotential(Fe=1).keys():
            self.assertIsInstance(k, Element)

    def test_math(self):
        fepot = ChemicalPotential({"Fe": 1})
        opot = ChemicalPotential({"O": 2.1})
        pots = ChemicalPotential({"Fe": 1, "O": 2.1})
        potsx2 = ChemicalPotential({"Fe": 2, "O": 4.2})
        feo2 = Composition("FeO2")

        # test get_energy()
        self.assertAlmostEqual(pots.get_energy(feo2), 5.2)
        self.assertAlmostEqual(fepot.get_energy(feo2, False), 1)
        self.assertRaises(ValueError, fepot.get_energy, feo2)

        # test multiplication
        self.assertRaises(NotImplementedError, lambda: (pots * pots))
        self.assertDictEqual(pots * 2, potsx2)
        self.assertDictEqual(2 * pots, potsx2)

        # test division
        self.assertDictEqual(potsx2 / 2, pots)
        self.assertRaises(NotImplementedError, lambda: (pots / pots))
        self.assertRaises(NotImplementedError, lambda: (pots / feo2))

        # test add/subtract
        self.assertDictEqual(pots + pots, potsx2)
        self.assertDictEqual(potsx2 - pots, pots)
        self.assertDictEqual(fepot + opot, pots)
        self.assertDictEqual(fepot - opot, pots - opot - opot)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


"""
Created on Jan 22, 2013

@author: Bharat Medasani
"""
import os
import sys
import unittest

from monty.os.path import which

from pymatgen.analysis.bond_valence import BVAnalyzer
from pymatgen.command_line.gulp_caller import (
    BuckinghamPotential,
    GulpCaller,
    GulpError,
    GulpIO,
    get_energy_buckingham,
    get_energy_relax_structure_buckingham,
    get_energy_tersoff,
)
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.util.testing import PymatgenTest

gulp_present = which("gulp") and os.environ.get("GULP_LIB") and ("win" not in sys.platform)
# disable gulp tests for now. Right now, it is compiled against libgfortran3, which is no longer supported in the new
# Ubuntu 20.04.
gulp_present = False


@unittest.skipIf(not gulp_present, "gulp not present.")
class GulpCallerTest(unittest.TestCase):
    def test_run(self):
        mgo_latt = [[4.212, 0, 0], [0, 4.212, 0], [0, 0, 4.212]]
        mgo_specie = ["Mg"] * 4 + ["O"] * 4
        mgo_frac_cord = [
            [0, 0, 0],
            [0.5, 0.5, 0],
            [0.5, 0, 0.5],
            [0, 0.5, 0.5],
            [0.5, 0, 0],
            [0, 0.5, 0],
            [0, 0, 0.5],
            [0.5, 0.5, 0.5],
        ]
        mgo_uc = Structure(mgo_latt, mgo_specie, mgo_frac_cord, True, True)
        gio = GulpIO()
        gin = gio.keyword_line("optimise", "conp")
        gin += gio.structure_lines(mgo_uc, symm_flg=False)
        # gin += self.gc.gulp_lib('catlow.lib')
        gin += "species\nMg    core  2.00000\nO core  0.86902\nO shel -2.86902\n"
        gin += "buck\n"
        gin += "Mg core O shel   946.627 0.31813  0.00000 0.0 10.0\n"
        gin += "O  shel O shel 22764.000 0.14900 27.87900 0.0 12.0\n"
        gin = gin
        gc = GulpCaller()

        """Some inherent checks are in the run_gulp function itself.
        They should be suffcient for raising errors."""
        gout = gc.run(gin)

    def test_decimal(self):
        struct = Structure.from_str(
            """Mg2 Al4 O8
        1.0
        5.003532 0.000000 2.888790
        1.667844 4.717375 2.888790
        0.000000 0.000000 5.777581
        O Mg Al
        8 2 4
        direct
        0.736371 0.736371 0.736371 O
        0.263629 0.263629 0.709114 O
        0.263629 0.709114 0.263629 O
        0.709114 0.263629 0.263629 O
        0.736371 0.290886 0.736371 O
        0.290886 0.736371 0.736371 O
        0.263629 0.263629 0.263629 O
        0.736371 0.736371 0.290886 O
        0.125000 0.125000 0.125000 Mg
        0.875000 0.875000 0.875000 Mg
        0.500000 0.500000 0.000000 Al
        0.500000 0.500000 0.500000 Al
        0.000000 0.500000 0.500000 Al
        0.500000 0.000000 0.500000 Al""",
            fmt="poscar",
        )

        bp = BuckinghamPotential(bush_lewis_flag="bush")
        gio = GulpIO()
        input = gio.buckingham_input(struct, ["relax conp"])
        caller = GulpCaller()
        gout = caller.run(input)


@unittest.skipIf(not gulp_present, "gulp not present.")
class GulpIOTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def setUp(self):
        p = Poscar.from_file(os.path.join(PymatgenTest.TEST_FILES_DIR, "POSCAR.Al12O18"), check_for_POTCAR=False)
        self.structure = p.structure
        self.gio = GulpIO()

    def test_keyword_line_with_correct_keywords(self):
        kw = ("defect", "property")
        inp_str = self.gio.keyword_line(*kw)
        for word in kw:
            self.assertIn(word, inp_str)

    def test_structure_lines_default_options(self):
        inp_str = self.gio.structure_lines(self.structure)
        self.assertIn("cell", inp_str)
        self.assertIn("frac", inp_str)
        self.assertIn("space", inp_str)

    def test_structure_lines_no_unitcell(self):
        inp_str = self.gio.structure_lines(self.structure, cell_flg=False)
        self.assertNotIn("cell", inp_str)

    def test_structure_lines_no_frac_coords(self):
        inp_str = self.gio.structure_lines(self.structure, cell_flg=False, frac_flg=False)
        self.assertNotIn("cell", inp_str)
        self.assertIn("cart", inp_str)

    @unittest.skip("Not Implemented yet")
    def test_specie_potential(self):
        pass

    @unittest.expectedFailure
    def test_library_line_explicit_path(self):
        gin = self.gio.library_line("/Users/mbkumar/Research/Defects/GulpExe/Libraries/catlow.lib")
        self.assertIn("lib", gin)

    def test_library_line_wrong_file(self):
        with self.assertRaises(GulpError):
            gin = self.gio.library_line("temp_to_fail.lib")

    def test_buckingham_potential(self):
        mgo_latt = [[4.212, 0, 0], [0, 4.212, 0], [0, 0, 4.212]]
        mgo_specie = ["Mg", "O"] * 4
        mgo_frac_cord = [
            [0, 0, 0],
            [0.5, 0, 0],
            [0.5, 0.5, 0],
            [0, 0.5, 0],
            [0.5, 0, 0.5],
            [0, 0, 0.5],
            [0, 0.5, 0.5],
            [0.5, 0.5, 0.5],
        ]
        mgo_uc = Structure(mgo_latt, mgo_specie, mgo_frac_cord, True, True)
        gin = self.gio.buckingham_potential(mgo_uc)
        self.assertIn("specie", gin)
        self.assertIn("buck", gin)
        self.assertIn("spring", gin)
        self.assertIn("Mg core", gin)
        self.assertIn("O  core", gin)
        self.assertIn("O  shel", gin)

        gin = self.gio.buckingham_potential(self.structure)
        self.assertIn("specie", gin)
        self.assertIn("buck", gin)
        self.assertIn("spring", gin)

    def test_buckingham_input(self):
        mgo_latt = [[4.212, 0, 0], [0, 4.212, 0], [0, 0, 4.212]]
        mgo_specie = ["Mg", "O"] * 4
        mgo_frac_cord = [
            [0, 0, 0],
            [0.5, 0, 0],
            [0.5, 0.5, 0],
            [0, 0.5, 0],
            [0.5, 0, 0.5],
            [0, 0, 0.5],
            [0, 0.5, 0.5],
            [0.5, 0.5, 0.5],
        ]
        mgo_uc = Structure(mgo_latt, mgo_specie, mgo_frac_cord, True, True)
        gin = self.gio.buckingham_input(mgo_uc, keywords=("optimise", "conp"))
        self.assertIn("optimise", gin)
        self.assertIn("cell", gin)
        self.assertIn("specie", gin)
        self.assertIn("buck", gin)
        self.assertIn("spring", gin)
        self.assertIn("Mg core", gin)
        self.assertIn("O  core", gin)
        self.assertIn("O  shel", gin)

    # Improve the test
    def test_tersoff_potential(self):
        mgo_latt = [[4.212, 0, 0], [0, 4.212, 0], [0, 0, 4.212]]
        mgo_specie = ["Mg", "O"] * 4
        mgo_frac_cord = [
            [0, 0, 0],
            [0.5, 0, 0],
            [0.5, 0.5, 0],
            [0, 0.5, 0],
            [0.5, 0, 0.5],
            [0, 0, 0.5],
            [0, 0.5, 0.5],
            [0.5, 0.5, 0.5],
        ]
        mgo_uc = Structure(mgo_latt, mgo_specie, mgo_frac_cord, True, True)
        gin = self.gio.tersoff_potential(mgo_uc)
        self.assertIn("specie", gin)
        self.assertIn("Mg core", gin)

    def test_get_energy(self):
        # Output string obtained from running GULP on a terminal
        out_str = """  Components of energy :
--------------------------------------------------------------------------------
  Interatomic potentials     =           5.61135426 eV
  Monopole - monopole (real) =          -4.34238722 eV
  Monopole - monopole (recip)=         -43.45344934 eV
  Monopole - monopole (total)=         -47.79583656 eV
--------------------------------------------------------------------------------
  Total lattice energy :
    Primitive unit cell      =         -42.18448230 eV
    Non-primitive unit cell  =        -168.73792920 eV
--------------------------------------------------------------------------------
  Total lattice energy (in kJmol-1):
    Primitive unit cell      =           -4070.1577 kJ/(mole unit cells)
    Non-primitive unit cell  =          -16280.6308 kJ/(mole unit cells)
--------------------------------------------------------------------------------
  Components of energy :

--------------------------------------------------------------------------------
  Interatomic potentials     =           6.79846039 eV
  Monopole - monopole (real) =          -4.45761741 eV
  Monopole - monopole (recip)=         -44.60653603 eV
  Monopole - monopole (total)=         -49.06415344 eV
--------------------------------------------------------------------------------
  Total lattice energy :
    Primitive unit cell      =         -42.26569304 eV
    Non-primitive unit cell  =        -169.06277218 eV
--------------------------------------------------------------------------------
  Total lattice energy (in kJmol-1):
    Primitive unit cell      =           -4077.9933 kJ/(mole unit cells)
    Non-primitive unit cell  =          -16311.9732 kJ/(mole unit cells)
--------------------------------------------------------------------------------"""
        energy = self.gio.get_energy(out_str)
        self.assertEqual(energy, -169.06277218)

    def test_get_relaxed_structure(self):
        # Output string obtained from running GULP on a terminal

        with open(os.path.join(PymatgenTest.TEST_FILES_DIR, "example21.gout"), "r") as fp:
            out_str = fp.read()
        struct = self.gio.get_relaxed_structure(out_str)
        self.assertIsInstance(struct, Structure)
        self.assertEqual(8, len(struct.sites))
        self.assertEqual(4.212, struct.lattice.a)
        self.assertEqual(90, struct.lattice.alpha)

    @unittest.skip("Test later")
    def test_tersoff_inpt(self):
        gin = self.gio.tersoff_input(self.structure)


@unittest.skipIf(not gulp_present, "gulp not present.")
class GlobalFunctionsTest(unittest.TestCase):
    def setUp(self):
        mgo_latt = [[4.212, 0, 0], [0, 4.212, 0], [0, 0, 4.212]]
        mgo_specie = ["Mg", "O"] * 4
        mgo_frac_cord = [
            [0, 0, 0],
            [0.5, 0, 0],
            [0.5, 0.5, 0],
            [0, 0.5, 0],
            [0.5, 0, 0.5],
            [0, 0, 0.5],
            [0, 0.5, 0.5],
            [0.5, 0.5, 0.5],
        ]
        self.mgo_uc = Structure(mgo_latt, mgo_specie, mgo_frac_cord, True, True)
        bv = BVAnalyzer()
        val = bv.get_valences(self.mgo_uc)
        el = [site.species_string for site in self.mgo_uc.sites]
        self.val_dict = dict(zip(el, val))

    def test_get_energy_tersoff(self):
        p = Poscar.from_file(os.path.join(PymatgenTest.TEST_FILES_DIR, "POSCAR.Al12O18"), check_for_POTCAR=False)
        structure = p.structure
        enrgy = get_energy_tersoff(structure)
        self.assertIsInstance(enrgy, float)

    def test_get_energy_buckingham(self):
        enrgy = get_energy_buckingham(self.mgo_uc)
        self.assertIsInstance(enrgy, float)
        # test with vacancy structure
        del self.mgo_uc[0]
        energy = get_energy_buckingham(
            self.mgo_uc,
            keywords=("qok", "optimise", "conp"),
            valence_dict=self.val_dict,
        )
        self.assertIsInstance(energy, float)

    def test_get_energy_relax_structure_buckingham(self):
        enrgy, struct = get_energy_relax_structure_buckingham(self.mgo_uc)
        self.assertIsInstance(enrgy, float)
        self.assertIsInstance(struct, Structure)
        site_len = len(struct.sites)
        self.assertEqual(site_len, len(self.mgo_uc.sites))


@unittest.skipIf(not gulp_present, "gulp not present.")
class BuckinghamPotentialLewisTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def setUp(self):
        self.bpl = BuckinghamPotential("lewis")

    def test_existing_element(self):
        self.assertIn("Sc_2+", self.bpl.pot_dict.keys())
        self.assertIn("Sc_2+", self.bpl.species_dict.keys())
        self.assertIn("O", self.bpl.pot_dict.keys())
        self.assertIn("O_core", self.bpl.species_dict.keys())
        self.assertIn("O_shel", self.bpl.species_dict.keys())

    def test_non_exisitng_element(self):
        self.assertNotIn("Li_1+", self.bpl.pot_dict.keys())
        self.assertNotIn("Li_1+", self.bpl.species_dict.keys())

    def test_element_different_valence(self):
        self.assertNotIn("Sc_4+", self.bpl.species_dict.keys())

    def test_values(self):
        self.assertNotEqual("", self.bpl.species_dict["Sc_2+"])
        self.assertNotEqual("", self.bpl.pot_dict["Sc_2+"])

    def test_spring(self):
        self.assertNotIn("Li", self.bpl.spring_dict.keys())
        self.assertNotEqual("", self.bpl.spring_dict["O"])


@unittest.skipIf(not gulp_present, "gulp not present.")
class BuckinghamPotentialBushTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def setUp(self):
        self.bpb = BuckinghamPotential("bush")

    def test_existing_element(self):
        self.assertIn("Li", self.bpb.pot_dict.keys())
        self.assertIn("Li", self.bpb.species_dict.keys())
        self.assertIn("O", self.bpb.pot_dict.keys())
        self.assertIn("O", self.bpb.species_dict.keys())

    def test_non_exisitng_element(self):
        self.assertNotIn("Mn", self.bpb.pot_dict.keys())
        self.assertNotIn("Mn", self.bpb.species_dict.keys())

    def test_element_different_valence(self):
        self.assertNotEqual(2, self.bpb.species_dict["Li"]["oxi"])

    def test_spring(self):
        self.assertEqual("", self.bpb.spring_dict["Li"])
        self.assertNotEqual("", self.bpb.spring_dict["O"])


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


__author__ = "Nils Edvin Richard Zimmermann"
__copyright__ = "Copyright 2016, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Nils Edvin Richard Zimmermann"
__email__ = "nils.e.r.zimmermann@gmail.com"
__date__ = "June 23, 2016"

import os
import unittest

from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.io.xr import Xr
from pymatgen.util.testing import PymatgenTest


class XrTest(unittest.TestCase):
    def setUp(self):
        filepath = os.path.join(PymatgenTest.TEST_FILES_DIR, "POSCAR")
        p = Poscar.from_file(filepath)
        self.xr = Xr(p.structure)

    def test_str(self):
        expected_string = """pymatgen   10.4118 6.0672 4.7595
90.000 90.000 90.000
24 0
0 Fe4 P4 O16
1 Fe 2.2773 4.5504 2.2601
2 Fe 2.9285 1.5168 4.6399
3 Fe 7.4832 4.5504 0.1196
4 Fe 8.1344 1.5168 2.4994
5 P 0.9851 1.5168 1.9906
6 P 4.2208 4.5504 4.3704
7 P 6.1910 1.5168 0.3891
8 P 9.4267 4.5504 2.7689
9 O 0.4516 4.5504 3.3656
10 O 1.0062 1.5168 3.5283
11 O 1.7253 0.2795 1.3583
12 O 1.7253 2.7541 1.3583
13 O 3.4806 3.3131 3.7380
14 O 3.4806 5.7876 3.7380
15 O 4.1997 4.5504 1.1486
16 O 4.7543 1.5168 0.9859
17 O 5.6575 4.5504 3.7736
18 O 6.2121 1.5168 3.6109
19 O 6.9312 0.2795 1.0215
20 O 6.9312 2.7541 1.0215
21 O 8.6864 3.3131 3.4012
22 O 8.6864 5.7876 3.4012
23 O 9.4055 4.5504 1.2312
24 O 9.9602 1.5168 1.3939
10.4118 0.0000 0.0000
0.0000 6.0672 0.0000
0.0000 0.0000 4.7595
10.4118 0.0000 0.0000
0.0000 6.0672 0.0000
0.0000 0.0000 4.7595"""
        self.assertEqual(str(self.xr), expected_string)

    def test_from_file(self):
        filename = os.path.join(PymatgenTest.TEST_FILES_DIR, "EDI.xr")
        xr = Xr.from_file(filename)
        self.assertIsInstance(xr.structure, Structure)
        xr2 = Xr.from_file(filename, use_cores=False)
        self.assertIsInstance(xr2.structure, Structure)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


import os
import unittest

from pymatgen.core.structure import Structure
from pymatgen.io.prismatic import Prismatic
from pymatgen.util.testing import PymatgenTest


class PrismaticTest(unittest.TestCase):
    def test_to_string(self):

        structure = Structure.from_file(os.path.join(PymatgenTest.TEST_FILES_DIR, "CuCl.cif"))
        prismatic = Prismatic(structure)
        prismatic_str = prismatic.to_string()
        self.assertTrue(
            prismatic_str.startswith(
                """Generated by pymatgen
6.52372159 6.52372159 6.52372159
29"""
            )
        )
        self.assertTrue(prismatic_str.endswith("-1"))



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


import unittest

from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.symmetry.kpath import KPathSeek
from pymatgen.util.testing import PymatgenTest

try:
    from seekpath import get_path  # type: ignore
except ImportError:
    get_path = None


class KPathSeekTest(PymatgenTest):
    @unittest.skipIf(get_path is None, "No seek path present.")
    def test_kpath_generation(self):
        triclinic = [1, 2]
        monoclinic = range(3, 16)
        orthorhombic = range(16, 75)
        tetragonal = range(75, 143)
        rhombohedral = range(143, 168)
        hexagonal = range(168, 195)
        cubic = range(195, 231)

        species = ["K", "La", "Ti"]
        coords = [[0.345, 5, 0.77298], [0.1345, 5.1, 0.77298], [0.7, 0.8, 0.9]]
        for i in range(230):
            sg_num = i + 1
            if sg_num in triclinic:
                lattice = Lattice(
                    [
                        [3.0233057319441246, 1, 0],
                        [0, 7.9850357844548681, 1],
                        [0, 1.2, 8.1136762279561818],
                    ]
                )
            elif sg_num in monoclinic:
                lattice = Lattice.monoclinic(2, 9, 1, 99)
            elif sg_num in orthorhombic:
                lattice = Lattice.orthorhombic(2, 9, 1)
            elif sg_num in tetragonal:
                lattice = Lattice.tetragonal(2, 9)
            elif sg_num in rhombohedral:
                lattice = Lattice.hexagonal(2, 95)
            elif sg_num in hexagonal:
                lattice = Lattice.hexagonal(2, 9)
            elif sg_num in cubic:
                lattice = Lattice.cubic(2)

            struct = Structure.from_spacegroup(sg_num, lattice, species, coords)
            kpath = KPathSeek(struct)  # Throws error if something doesn't work, causing test to fail.
            kpoints = kpath.get_kpoints()  # noqa: F841

    @unittest.skipIf(get_path is None, "No seek path present.")
    def test_kpath_acentered(self):
        species = ["K", "La", "Ti"]
        coords = [[0.345, 5, 0.77298], [0.1345, 5.1, 0.77298], [0.7, 0.8, 0.9]]
        lattice = Lattice.orthorhombic(2, 9, 1)
        struct = Structure.from_spacegroup(38, lattice, species, coords)
        kpath = KPathSeek(struct)

        kpoints = kpath._kpath["kpoints"]
        labels = list(kpoints.keys())

        self.assertEqual(
            sorted(labels),
            sorted(
                [
                    "B_0",
                    "B_2",
                    "DELTA_0",
                    "F_0",
                    "GAMMA",
                    "G_0",
                    "G_2",
                    "R",
                    "R_2",
                    "S",
                    "T",
                    "T_2",
                    "Y",
                    "Z",
                    "Z_2",
                ]
            ),
        )

        self.assertAlmostEqual(kpoints["GAMMA"][0], 0.0)
        self.assertAlmostEqual(kpoints["GAMMA"][1], 0.0)
        self.assertAlmostEqual(kpoints["GAMMA"][2], 0.0)

        self.assertAlmostEqual(kpoints["Y"][0], 0.5)
        self.assertAlmostEqual(kpoints["Y"][1], 0.5)
        self.assertAlmostEqual(kpoints["Y"][2], 0.0)

        self.assertAlmostEqual(kpoints["T"][0], 0.5)
        self.assertAlmostEqual(kpoints["T"][1], 0.5)
        self.assertAlmostEqual(kpoints["T"][2], 0.5)

        self.assertAlmostEqual(kpoints["T_2"][0], 0.5)
        self.assertAlmostEqual(kpoints["T_2"][1], 0.5)
        self.assertAlmostEqual(kpoints["T_2"][2], -0.5)

        self.assertAlmostEqual(kpoints["Z"][0], 0.0)
        self.assertAlmostEqual(kpoints["Z"][1], 0.0)
        self.assertAlmostEqual(kpoints["Z"][2], 0.5)

        self.assertAlmostEqual(kpoints["Z_2"][0], 0.0)
        self.assertAlmostEqual(kpoints["Z_2"][1], 0.0)
        self.assertAlmostEqual(kpoints["Z_2"][2], -0.5)

        self.assertAlmostEqual(kpoints["S"][0], 0.0)
        self.assertAlmostEqual(kpoints["S"][1], 0.5)
        self.assertAlmostEqual(kpoints["S"][2], 0.0)

        self.assertAlmostEqual(kpoints["R"][0], 0.0)
        self.assertAlmostEqual(kpoints["R"][1], 0.5)
        self.assertAlmostEqual(kpoints["R"][2], 0.5)

        self.assertAlmostEqual(kpoints["R_2"][0], 0.0)
        self.assertAlmostEqual(kpoints["R_2"][1], 0.5)
        self.assertAlmostEqual(kpoints["R_2"][2], -0.5)

        self.assertAlmostEqual(kpoints["DELTA_0"][0], -0.25308641975308643)
        self.assertAlmostEqual(kpoints["DELTA_0"][1], 0.25308641975308643)
        self.assertAlmostEqual(kpoints["DELTA_0"][2], 0.0)

        self.assertAlmostEqual(kpoints["F_0"][0], 0.25308641975308643)
        self.assertAlmostEqual(kpoints["F_0"][1], 0.7469135802469136)
        self.assertAlmostEqual(kpoints["F_0"][2], 0.0)

        self.assertAlmostEqual(kpoints["B_0"][0], -0.25308641975308643)
        self.assertAlmostEqual(kpoints["B_0"][1], 0.25308641975308643)
        self.assertAlmostEqual(kpoints["B_0"][2], 0.5)

        self.assertAlmostEqual(kpoints["B_2"][0], -0.25308641975308643)
        self.assertAlmostEqual(kpoints["B_2"][1], 0.25308641975308643)
        self.assertAlmostEqual(kpoints["B_2"][2], -0.5)

        self.assertAlmostEqual(kpoints["G_0"][0], 0.25308641975308643)
        self.assertAlmostEqual(kpoints["G_0"][1], 0.7469135802469136)
        self.assertAlmostEqual(kpoints["G_0"][2], 0.5)

        self.assertAlmostEqual(kpoints["G_2"][0], 0.25308641975308643)
        self.assertAlmostEqual(kpoints["G_2"][1], 0.7469135802469136)
        self.assertAlmostEqual(kpoints["G_2"][2], -0.5)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

import numpy as np

from pymatgen.core.lattice import Lattice
from pymatgen.symmetry.groups import SpaceGroup
from pymatgen.symmetry.maggroups import MagneticSpaceGroup
from pymatgen.util.testing import PymatgenTest

__author__ = "Matthew Horton"
__copyright__ = "Copyright 2017, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Matthew Horton"
__email__ = "mkhorton@lbl.gov"
__status__ = "Beta"
__date__ = "Feb 2017"


class MagneticSpaceGroupTest(PymatgenTest):
    def setUp(self):
        self.msg_1 = MagneticSpaceGroup([70, 530])
        self.msg_2 = MagneticSpaceGroup([62, 448])
        self.msg_3 = MagneticSpaceGroup([20, 37])
        self.msg_4 = MagneticSpaceGroup([2, 7], "c,1/4a+1/4b,-1/2a+1/2b;0,0,0")

    def test_init(self):
        # test init with the following space group:
        # 71.538 (BNS number), I_cmmm (BNS label)
        # 65.10.554 (same space group as above, OG number), C_Immm (OG label)
        msg_from_bns_1 = MagneticSpaceGroup("I_cmmm")
        msg_from_bns_2 = MagneticSpaceGroup([71, 538])
        msg_from_og_1 = MagneticSpaceGroup.from_og("C_Immm")
        msg_from_og_2 = MagneticSpaceGroup.from_og([65, 10, 554])
        self.assertEqual(msg_from_bns_1, msg_from_bns_2)
        self.assertEqual(msg_from_og_1, msg_from_og_2)
        self.assertEqual(msg_from_bns_1, msg_from_og_1)

    def test_crystal_system(self):
        self.assertEqual(self.msg_1.crystal_system, "orthorhombic")
        self.assertEqual(self.msg_2.crystal_system, "orthorhombic")
        self.assertEqual(self.msg_3.crystal_system, "orthorhombic")

    def test_sg_symbol(self):
        self.assertEqual(self.msg_1.sg_symbol, "Fd'd'd")
        self.assertEqual(self.msg_2.sg_symbol, "Pn'ma'")
        self.assertEqual(self.msg_3.sg_symbol, "C_A222_1")

    def test_is_compatible(self):
        cubic = Lattice.cubic(1)
        hexagonal = Lattice.hexagonal(1, 2)
        rhom = Lattice.rhombohedral(3, 80)
        tet = Lattice.tetragonal(1, 2)
        ortho = Lattice.orthorhombic(1, 2, 3)
        msg = MagneticSpaceGroup("Fm-3m")
        self.assertTrue(msg.is_compatible(cubic))
        self.assertFalse(msg.is_compatible(hexagonal))
        msg = MagneticSpaceGroup("Pnma")
        self.assertTrue(msg.is_compatible(cubic))
        self.assertTrue(msg.is_compatible(tet))
        self.assertTrue(msg.is_compatible(ortho))
        self.assertFalse(msg.is_compatible(rhom))
        self.assertFalse(msg.is_compatible(hexagonal))
        msg = MagneticSpaceGroup("P2/c")
        self.assertTrue(msg.is_compatible(cubic))
        self.assertTrue(msg.is_compatible(tet))
        self.assertTrue(msg.is_compatible(ortho))
        self.assertFalse(msg.is_compatible(rhom))
        self.assertFalse(msg.is_compatible(hexagonal))
        msg = MagneticSpaceGroup("P-1")
        self.assertTrue(msg.is_compatible(cubic))
        self.assertTrue(msg.is_compatible(tet))
        self.assertTrue(msg.is_compatible(ortho))
        self.assertTrue(msg.is_compatible(rhom))
        self.assertTrue(msg.is_compatible(hexagonal))

    def test_symmetry_ops(self):

        msg_1_symmops = "\n".join([str(op) for op in self.msg_1.symmetry_ops])
        msg_1_symmops_ref = """x, y, z, +1
-x+3/4, -y+3/4, z, +1
-x, -y, -z, +1
x+1/4, y+1/4, -z, +1
x, -y+3/4, -z+3/4, -1
-x+3/4, y, -z+3/4, -1
-x, y+1/4, z+1/4, -1
x+1/4, -y, z+1/4, -1
x, y+1/2, z+1/2, +1
-x+3/4, -y+5/4, z+1/2, +1
-x, -y+1/2, -z+1/2, +1
x+1/4, y+3/4, -z+1/2, +1
x, -y+5/4, -z+5/4, -1
-x+3/4, y+1/2, -z+5/4, -1
-x, y+3/4, z+3/4, -1
x+1/4, -y+1/2, z+3/4, -1
x+1/2, y, z+1/2, +1
-x+5/4, -y+3/4, z+1/2, +1
-x+1/2, -y, -z+1/2, +1
x+3/4, y+1/4, -z+1/2, +1
x+1/2, -y+3/4, -z+5/4, -1
-x+5/4, y, -z+5/4, -1
-x+1/2, y+1/4, z+3/4, -1
x+3/4, -y, z+3/4, -1
x+1/2, y+1/2, z, +1
-x+5/4, -y+5/4, z, +1
-x+1/2, -y+1/2, -z, +1
x+3/4, y+3/4, -z, +1
x+1/2, -y+5/4, -z+3/4, -1
-x+5/4, y+1/2, -z+3/4, -1
-x+1/2, y+3/4, z+1/4, -1
x+3/4, -y+1/2, z+1/4, -1"""

        msg_2_symmops = "\n".join([str(op) for op in self.msg_2.symmetry_ops])
        msg_2_symmops_ref = """x, y, z, +1
-x, y+1/2, -z, +1
-x, -y, -z, +1
x, -y+1/2, z, +1
x+1/2, -y+1/2, -z+1/2, -1
-x+1/2, -y, z+1/2, -1
-x+1/2, y+1/2, z+1/2, -1
x+1/2, y, -z+1/2, -1"""
        self.assertStrContentEqual(msg_2_symmops, msg_2_symmops_ref)

        msg_3_symmops = "\n".join([str(op) for op in self.msg_3.symmetry_ops])
        msg_3_symmops_ref = """x, y, z, +1
x, -y, -z, +1
-x, y, -z+1/2, +1
-x, -y, z+1/2, +1
x, y+1/2, z+1/2, -1
x+1/2, -y, -z+1/2, -1
-x+1/2, y, -z, -1
-x+1/2, -y, z, -1
x+1/2, y+1/2, z, +1
x+1/2, -y+1/2, -z, +1
-x+1/2, y+1/2, -z+1/2, +1
-x+1/2, -y+1/2, z+1/2, +1
x+1/2, y, z+1/2, -1
x, -y+1/2, -z+1/2, -1
-x, y+1/2, -z, -1
-x, -y+1/2, z, -1"""
        self.assertEqual(msg_3_symmops, msg_3_symmops_ref)

        msg_4_symmops = "\n".join([str(op) for op in self.msg_4.symmetry_ops])
        msg_4_symmops_ref = """x, y, z, +1
-x, -y, -z, +1
x+1/2, y, z, -1
-x+1/2, -y, -z, -1"""
        self.assertEqual(msg_4_symmops, msg_4_symmops_ref)

    def test_equivalence_to_spacegroup(self):

        # first 230 magnetic space groups have same symmetry operations
        # as normal space groups, so should give same orbits

        labels = ["Fm-3m", "Pnma", "P2/c", "P-1"]

        points = [[0, 0, 0], [0.5, 0, 0], [0.11, 0.22, 0.33]]

        for label in labels:
            sg = SpaceGroup(label)
            msg = MagneticSpaceGroup(label)
            self.assertEqual(sg.crystal_system, msg.crystal_system)
            for p in points:
                pp_sg = np.array(sg.get_orbit(p))
                pp_msg = np.array(msg.get_orbit(p, 0)[0])  # discarding magnetic moment information
                pp_sg = pp_sg[np.lexsort(np.transpose(pp_sg)[::-1])]  # sorting arrays so we can compare them
                pp_msg = pp_msg[np.lexsort(np.transpose(pp_msg)[::-1])]
                self.assertTrue(np.allclose(pp_sg, pp_msg))

    def test_str(self):

        msg = MagneticSpaceGroup([4, 11])

        ref_string = """BNS: 4.11 P_b2_1
Operators: (1|0,0,0) (2y|0,1/2,0) (1|0,1/2,0)' (2y|0,0,0)'
Wyckoff Positions:
4e  (x,y,z;mx,my,mz) (-x,y+1/2,-z;-mx,my,-mz) (x,y+1/2,z;-mx,-my,-mz)
    (-x,y,-z;mx,-my,mz)
2d  (1/2,y,1/2;mx,0,mz) (1/2,y+1/2,1/2;-mx,0,-mz)
2c  (1/2,y,0;mx,0,mz) (1/2,y+1/2,0;-mx,0,-mz)
2b  (0,y,1/2;mx,0,mz) (0,y+1/2,1/2;-mx,0,-mz)
2a  (0,y,0;mx,0,mz) (0,y+1/2,0;-mx,0,-mz)
Alternative OG setting exists for this space group."""

        ref_string_all = """BNS: 4.11 P_b2_1		OG: 3.7.14 P_2b2'
OG-BNS Transform: (a,2b,c;0,0,0)
Operators (BNS): (1|0,0,0) (2y|0,1/2,0) (1|0,1/2,0)' (2y|0,0,0)'
Wyckoff Positions (BNS):
4e  (x,y,z;mx,my,mz) (-x,y+1/2,-z;-mx,my,-mz) (x,y+1/2,z;-mx,-my,-mz)
    (-x,y,-z;mx,-my,mz)
2d  (1/2,y,1/2;mx,0,mz) (1/2,y+1/2,1/2;-mx,0,-mz)
2c  (1/2,y,0;mx,0,mz) (1/2,y+1/2,0;-mx,0,-mz)
2b  (0,y,1/2;mx,0,mz) (0,y+1/2,1/2;-mx,0,-mz)
2a  (0,y,0;mx,0,mz) (0,y+1/2,0;-mx,0,-mz)
Operators (OG): (1|0,0,0) (2y|0,1,0) (1|0,1,0)' (2y|0,0,0)'
Wyckoff Positions (OG): (1,0,0)+ (0,2,0)+ (0,0,1)+
4e  (x,y,z;mx,my,mz) (-x,y+1,-z;-mx,my,-mz) (x,y+1,z;-mx,-my,-mz)
    (-x,y,-z;mx,-my,mz)
2d  (1/2,y,1/2;mx,0,mz) (-1/2,y+1,-1/2;-mx,0,-mz)
2c  (1/2,y,0;mx,0,mz) (-1/2,y+1,0;-mx,0,-mz)
2b  (0,y,1/2;mx,0,mz) (0,y+1,-1/2;-mx,0,-mz)
2a  (0,y,0;mx,0,mz) (0,y+1,0;-mx,0,-mz)"""

        self.assertStrContentEqual(str(msg), ref_string)
        self.assertStrContentEqual(msg.data_str(), ref_string_all)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


import itertools

import numpy as np

from pymatgen.core.lattice import Lattice, get_points_in_spheres
from pymatgen.core.operations import SymmOp
from pymatgen.util.testing import PymatgenTest


class LatticeTestCase(PymatgenTest):
    def setUp(self):
        self.lattice = Lattice.cubic(10.0)
        self.cubic = self.lattice
        self.tetragonal = Lattice.tetragonal(10, 20)
        self.orthorhombic = Lattice.orthorhombic(10, 20, 30)
        self.monoclinic = Lattice.monoclinic(10, 20, 30, 66)
        self.hexagonal = Lattice.hexagonal(10, 20)
        self.rhombohedral = Lattice.rhombohedral(10, 77)

        family_names = [
            "cubic",
            "tetragonal",
            "orthorhombic",
            "monoclinic",
            "hexagonal",
            "rhombohedral",
        ]

        self.families = {}
        for name in family_names:
            self.families[name] = getattr(self, name)

    def test_format(self):
        self.assertEqual(
            "[[10.000, 0.000, 0.000], [0.000, 10.000, 0.000], [0.000, 0.000, 10.000]]",
            format(self.lattice, ".3fl"),
        )
        self.assertEqual(
            """10.000 0.000 0.000
0.000 10.000 0.000
0.000 0.000 10.000""",
            format(self.lattice, ".3f"),
        )
        self.assertEqual("{10.0, 10.0, 10.0, 90.0, 90.0, 90.0}", format(self.lattice, ".1fp"))

    def test_init(self):
        a = 9.026
        lattice = Lattice.cubic(a)
        self.assertIsNotNone(lattice, "Initialization from new_cubic failed")
        lattice2 = Lattice([[a, 0, 0], [0, a, 0], [0, 0, a]])
        for i in range(0, 3):
            for j in range(0, 3):
                self.assertAlmostEqual(
                    lattice.matrix[i][j],
                    lattice2.matrix[i][j],
                    5,
                    "Inconsistent matrix from two inits!",
                )

    def test_copy(self):
        cubic_copy = self.cubic.copy()
        self.assertTrue(cubic_copy == self.cubic)
        self.assertFalse(cubic_copy._matrix is self.cubic._matrix)

    def test_get_cartesian_or_frac_coord(self):
        coord = self.lattice.get_cartesian_coords([0.15, 0.3, 0.4])
        self.assertArrayAlmostEqual(coord, [1.5, 3.0, 4.0])
        self.assertArrayAlmostEqual(
            self.tetragonal.get_fractional_coords([12.12312, 45.2134, 1.3434]),
            [1.212312, 4.52134, 0.06717],
        )

        # Random testing that get_cart and get_frac coords reverses each other.
        rand_coord = np.random.random_sample(3)
        coord = self.tetragonal.get_cartesian_coords(rand_coord)
        fcoord = self.tetragonal.get_fractional_coords(coord)
        self.assertArrayAlmostEqual(fcoord, rand_coord)

    def test_get_vector_along_lattice_directions(self):
        lattice_mat = np.array([[0.5, 0.0, 0.0], [0.5, np.sqrt(3) / 2.0, 0.0], [0.0, 0.0, 1.0]])
        lattice = Lattice(lattice_mat)
        cart_coord = np.array([0.5, np.sqrt(3) / 4.0, 0.5])
        latt_coord = np.array([0.25, 0.5, 0.5])
        from_direct = lattice.get_fractional_coords(cart_coord) * lattice.lengths
        self.assertArrayAlmostEqual(lattice.get_vector_along_lattice_directions(cart_coord), from_direct)
        self.assertArrayAlmostEqual(lattice.get_vector_along_lattice_directions(cart_coord), latt_coord)
        self.assertArrayEqual(
            lattice.get_vector_along_lattice_directions(cart_coord).shape,
            [
                3,
            ],
        )
        self.assertArrayEqual(
            lattice.get_vector_along_lattice_directions(cart_coord.reshape([1, 3])).shape,
            [1, 3],
        )

    def test_d_hkl(self):
        cubic_copy = self.cubic.copy()
        hkl = (1, 2, 3)
        dhkl = ((hkl[0] ** 2 + hkl[1] ** 2 + hkl[2] ** 2) / (cubic_copy.a ** 2)) ** (-1 / 2)
        self.assertEqual(dhkl, cubic_copy.d_hkl(hkl))

    def test_reciprocal_lattice(self):
        recip_latt = self.lattice.reciprocal_lattice
        self.assertArrayAlmostEqual(recip_latt.matrix, 0.628319 * np.eye(3), 5)
        self.assertArrayAlmostEqual(
            self.tetragonal.reciprocal_lattice.matrix,
            [[0.628319, 0.0, 0.0], [0.0, 0.628319, 0], [0.0, 0.0, 0.3141590]],
            5,
        )

        # Test the crystallographic version.
        recip_latt_xtal = self.lattice.reciprocal_lattice_crystallographic
        self.assertArrayAlmostEqual(recip_latt.matrix, recip_latt_xtal.matrix * 2 * np.pi, 5)

    def test_static_methods(self):
        lengths_c = [3.840198, 3.84019885, 3.8401976]
        angles_c = [119.99998575, 90, 60.00000728]
        mat_c = [
            [3.840198, 0.000000, 0.0000],
            [1.920099, 3.325710, 0.000000],
            [0.000000, -2.217138, 3.135509],
        ]
        # should give the lengths and angles above
        newlatt = Lattice(mat_c)
        lengths = newlatt.lengths
        angles = newlatt.angles
        for i in range(0, 3):
            self.assertAlmostEqual(lengths[i], lengths_c[i], 5, "Lengths incorrect!")
            self.assertAlmostEqual(angles[i], angles_c[i], 5, "Angles incorrect!")
        latt = Lattice.from_parameters(*lengths, *angles)
        lengths = latt.lengths
        angles = latt.angles
        for i in range(0, 3):
            self.assertAlmostEqual(lengths[i], lengths_c[i], 5, "Lengths incorrect!")
            self.assertAlmostEqual(angles[i], angles_c[i], 5, "Angles incorrect!")

    def test_attributes(self):
        """docstring for test_attributes"""
        lattice = Lattice.cubic(10.0)
        self.assertEqual(lattice.a, 10.0)
        self.assertEqual(lattice.b, 10.0)
        self.assertEqual(lattice.c, 10.0)
        self.assertAlmostEqual(lattice.volume, 1000.0)
        xyz = lattice.get_cartesian_coords([0.25, 0.35, 0.45])
        self.assertEqual(xyz[0], 2.5)
        self.assertEqual(xyz[1], 3.5)
        self.assertEqual(xyz[2], 4.5)

    def test_lattice_matrices(self):
        """
        If alpha == 90 and beta == 90, two matricies are identical.
        """

        def _identical(a, b, c, alpha, beta, gamma):
            mat1 = Lattice.from_parameters(a, b, c, alpha, beta, gamma, False).matrix
            mat2 = Lattice.from_parameters(a, b, c, alpha, beta, gamma, True).matrix
            # self.assertArrayAlmostEqual(mat1, mat2)
            return ((mat1 - mat2) ** 2).sum() < 1e-6

        self.assertTrue(_identical(2, 3, 4, 90, 90, 90))
        self.assertTrue(_identical(2, 3, 4, 90, 90, 80))
        self.assertTrue(_identical(2, 3, 4, 90, 90, 100))

        self.assertFalse(_identical(2, 3, 4, 100, 90, 90))
        self.assertFalse(_identical(2, 3, 4, 90, 100, 90))
        self.assertFalse(_identical(2, 3, 4, 100, 100, 100))

    def test_get_lll_reduced_lattice(self):
        lattice = Lattice([1.0, 1, 1, -1.0, 0, 2, 3.0, 5, 6])
        reduced_latt = lattice.get_lll_reduced_lattice()

        expected_ans = Lattice(np.array([0.0, 1.0, 0.0, 1.0, 0.0, 1.0, -2.0, 0.0, 1.0]).reshape((3, 3)))
        self.assertAlmostEqual(np.linalg.det(np.linalg.solve(expected_ans.matrix, reduced_latt.matrix)), 1)
        self.assertArrayAlmostEqual(sorted(reduced_latt.abc), sorted(expected_ans.abc))
        self.assertAlmostEqual(reduced_latt.volume, lattice.volume)
        latt = [
            7.164750,
            2.481942,
            0.000000,
            -4.298850,
            2.481942,
            0.000000,
            0.000000,
            0.000000,
            14.253000,
        ]
        expected_ans = Lattice(
            np.array(
                [
                    -4.298850,
                    2.481942,
                    0.000000,
                    2.865900,
                    4.963884,
                    0.000000,
                    0.000000,
                    0.000000,
                    14.253000,
                ]
            )
        )
        reduced_latt = Lattice(latt).get_lll_reduced_lattice()
        self.assertAlmostEqual(np.linalg.det(np.linalg.solve(expected_ans.matrix, reduced_latt.matrix)), 1)
        self.assertArrayAlmostEqual(sorted(reduced_latt.abc), sorted(expected_ans.abc))

        expected_ans = Lattice([0.0, 10.0, 10.0, 10.0, 10.0, 0.0, 30.0, -30.0, 40.0])

        lattice = np.array([100.0, 0.0, 10.0, 10.0, 10.0, 20.0, 10.0, 10.0, 10.0])
        lattice = lattice.reshape(3, 3)
        lattice = Lattice(lattice.T)
        reduced_latt = lattice.get_lll_reduced_lattice()
        self.assertAlmostEqual(np.linalg.det(np.linalg.solve(expected_ans.matrix, reduced_latt.matrix)), 1)
        self.assertArrayAlmostEqual(sorted(reduced_latt.abc), sorted(expected_ans.abc))

        random_latt = Lattice(np.random.random((3, 3)))
        if np.linalg.det(random_latt.matrix) > 1e-8:
            reduced_random_latt = random_latt.get_lll_reduced_lattice()
            self.assertAlmostEqual(reduced_random_latt.volume, random_latt.volume)

    def test_get_niggli_reduced_lattice(self):
        latt = Lattice.from_parameters(3, 5.196, 2, 103 + 55 / 60, 109 + 28 / 60, 134 + 53 / 60)
        reduced_cell = latt.get_niggli_reduced_lattice()
        abc = reduced_cell.lengths
        angles = reduced_cell.angles
        self.assertAlmostEqual(abc[0], 2, 3)
        self.assertAlmostEqual(abc[1], 3, 3)
        self.assertAlmostEqual(abc[2], 3, 3)
        self.assertAlmostEqual(angles[0], 116.382855225, 3)
        self.assertAlmostEqual(angles[1], 94.769790287999996, 3)
        self.assertAlmostEqual(angles[2], 109.466666667, 3)

        mat = [[5.0, 0, 0], [0, 5.0, 0], [5.0, 0, 5.0]]
        latt = Lattice(np.dot([[1, 1, 1], [1, 1, 0], [0, 1, 1]], mat))
        reduced_cell = latt.get_niggli_reduced_lattice()
        abc = reduced_cell.lengths
        angles = reduced_cell.angles
        for l in abc:
            self.assertAlmostEqual(l, 5, 3)
        for a in angles:
            self.assertAlmostEqual(a, 90, 3)

        latt = Lattice(
            [
                1.432950,
                0.827314,
                4.751000,
                -1.432950,
                0.827314,
                4.751000,
                0.0,
                -1.654628,
                4.751000,
            ]
        )
        # ans = [[-1.432950, -2.481942, 0.0],
        #       [-2.8659, 0.0, 0.0],
        #       [-1.432950, -0.827314, -4.751000]]
        ans = [
            [-1.43295, -2.481942, 0.0],
            [-2.8659, 0.0, 0.0],
            [-1.43295, -0.827314, -4.751],
        ]
        self.assertArrayAlmostEqual(latt.get_niggli_reduced_lattice().matrix, ans)

        latt = Lattice.from_parameters(7.365450, 6.199506, 5.353878, 75.542191, 81.181757, 156.396627)
        ans = [
            [2.578932, 0.826965, 0.000000],
            [-0.831059, 2.067413, 1.547813],
            [-0.458407, -2.480895, 1.129126],
        ]
        self.assertArrayAlmostEqual(latt.get_niggli_reduced_lattice().matrix, np.array(ans), 5)

    def test_find_mapping(self):
        m = np.array([[0.1, 0.2, 0.3], [-0.1, 0.2, 0.7], [0.6, 0.9, 0.2]])
        latt = Lattice(m)

        op = SymmOp.from_origin_axis_angle([0, 0, 0], [2, 3, 3], 35)
        rot = op.rotation_matrix
        scale = np.array([[1, 1, 0], [0, 1, 0], [0, 0, 1]])

        latt2 = Lattice(np.dot(rot, np.dot(scale, m).T).T)
        (aligned_out, rot_out, scale_out) = latt2.find_mapping(latt)
        self.assertAlmostEqual(abs(np.linalg.det(rot)), 1)

        rotated = SymmOp.from_rotation_and_translation(rot_out).operate_multi(latt.matrix)

        self.assertArrayAlmostEqual(rotated, aligned_out.matrix)
        self.assertArrayAlmostEqual(np.dot(scale_out, latt2.matrix), aligned_out.matrix)
        self.assertArrayAlmostEqual(aligned_out.parameters, latt.parameters)
        self.assertFalse(np.allclose(aligned_out.parameters, latt2.parameters))

    def test_find_all_mappings(self):
        m = np.array([[0.1, 0.2, 0.3], [-0.1, 0.2, 0.7], [0.6, 0.9, 0.2]])
        latt = Lattice(m)

        op = SymmOp.from_origin_axis_angle([0, 0, 0], [2, -1, 3], 40)
        rot = op.rotation_matrix
        scale = np.array([[0, 2, 0], [1, 1, 0], [0, 0, 1]])

        latt2 = Lattice(np.dot(rot, np.dot(scale, m).T).T)

        for (aligned_out, rot_out, scale_out) in latt.find_all_mappings(latt2):
            self.assertArrayAlmostEqual(np.inner(latt2.matrix, rot_out), aligned_out.matrix, 5)
            self.assertArrayAlmostEqual(np.dot(scale_out, latt.matrix), aligned_out.matrix)
            self.assertArrayAlmostEqual(aligned_out.parameters, latt2.parameters)
            self.assertFalse(np.allclose(aligned_out.parameters, latt.parameters))

        latt = Lattice.orthorhombic(9, 9, 5)
        self.assertEqual(len(list(latt.find_all_mappings(latt))), 16)

        # catch the singular matrix error
        latt = Lattice.from_parameters(1, 1, 1, 10, 10, 10)
        for l, _, _ in latt.find_all_mappings(latt, ltol=0.05, atol=11):
            self.assertTrue(isinstance(l, Lattice))

    def test_mapping_symmetry(self):
        l = Lattice.cubic(1)
        l2 = Lattice.orthorhombic(1.1001, 1, 1)
        self.assertEqual(l.find_mapping(l2, ltol=0.1), None)
        self.assertEqual(l2.find_mapping(l, ltol=0.1), None)
        l2 = Lattice.orthorhombic(1.0999, 1, 1)
        self.assertNotEqual(l2.find_mapping(l, ltol=0.1), None)
        self.assertNotEqual(l.find_mapping(l2, ltol=0.1), None)

    def test_to_from_dict(self):
        d = self.tetragonal.as_dict()
        t = Lattice.from_dict(d)
        for i in range(3):
            self.assertEqual(t.abc[i], self.tetragonal.abc[i])
            self.assertEqual(t.angles[i], self.tetragonal.angles[i])
        # Make sure old style dicts work.
        d = self.tetragonal.as_dict(verbosity=1)
        del d["matrix"]
        t = Lattice.from_dict(d)
        for i in range(3):
            self.assertEqual(t.abc[i], self.tetragonal.abc[i])
            self.assertEqual(t.angles[i], self.tetragonal.angles[i])

    def test_scale(self):
        new_volume = 10
        for (family_name, lattice) in self.families.items():
            new_lattice = lattice.scale(new_volume)
            self.assertAlmostEqual(new_lattice.volume, new_volume)
            self.assertArrayAlmostEqual(new_lattice.angles, lattice.angles)

    def test_get_wigner_seitz_cell(self):
        ws_cell = Lattice([[10, 0, 0], [0, 5, 0], [0, 0, 1]]).get_wigner_seitz_cell()
        self.assertEqual(6, len(ws_cell))
        for l in ws_cell[3]:
            self.assertEqual([abs(i) for i in l], [5.0, 2.5, 0.5])

    def test_dot_and_norm(self):
        frac_basis = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

        for family_name, lattice in self.families.items():
            # print(family_name)
            self.assertArrayEqual(lattice.norm(lattice.matrix, frac_coords=False), lattice.abc)
            self.assertArrayEqual(lattice.norm(frac_basis), lattice.abc)
            for (i, vec) in enumerate(frac_basis):
                length = lattice.norm(vec)
                self.assertArrayEqual(length[0], lattice.abc[i])
                # We always get a ndarray.
                self.assertTrue(hasattr(length, "shape"))

        # Passing complex arrays should raise TypeError
        with self.assertRaises(TypeError):
            lattice.norm(np.zeros(3, dtype=np.complex128))

        # Cannot reshape the second argument.
        with self.assertRaises(ValueError):
            lattice.dot(np.zeros(6), np.zeros(8))

        # Passing vectors of different length is invalid.
        with self.assertRaises(ValueError):
            lattice.dot(np.zeros(3), np.zeros(6))

    def test_get_points_in_sphere(self):
        # This is a non-niggli representation of a cubic lattice
        latt = Lattice([[1, 5, 0], [0, 1, 0], [5, 0, 1]])
        # evenly spaced points array between 0 and 1
        pts = np.array(list(itertools.product(range(5), repeat=3))) / 5
        pts = latt.get_fractional_coords(pts)

        # Test getting neighbors within 1 neighbor distance of the origin
        fcoords, dists, inds, images = latt.get_points_in_sphere(pts, [0, 0, 0], 0.20001, zip_results=False)
        self.assertEqual(len(fcoords), 7)  # There are 7 neighbors
        self.assertEqual(np.isclose(dists, 0.2).sum(), 6)  # 6 are at 0.2
        self.assertEqual(np.isclose(dists, 0).sum(), 1)  # 1 is at 0
        self.assertEqual(len(set(inds)), 7)  # They have unique indices
        self.assertArrayEqual(images[np.isclose(dists, 0)], [[0, 0, 0]])

        # More complicated case, using the zip output
        result = latt.get_points_in_sphere(pts, [0.5, 0.5, 0.5], 1.0001)
        self.assertEqual(len(result), 552)
        self.assertEqual(len(result[0]), 4)  # coords, dists, ind, supercell

    def test_get_all_distances(self):
        fcoords = np.array(
            [
                [0.3, 0.3, 0.5],
                [0.1, 0.1, 0.3],
                [0.9, 0.9, 0.8],
                [0.1, 0.0, 0.5],
                [0.9, 0.7, 0.0],
            ]
        )
        lattice = Lattice.from_parameters(8, 8, 4, 90, 76, 58)
        expected = np.array(
            [
                [0.000, 3.015, 4.072, 3.519, 3.245],
                [3.015, 0.000, 3.207, 1.131, 4.453],
                [4.072, 3.207, 0.000, 2.251, 1.788],
                [3.519, 1.131, 2.251, 0.000, 3.852],
                [3.245, 4.453, 1.788, 3.852, 0.000],
            ]
        )
        output = lattice.get_all_distances(fcoords, fcoords)
        self.assertArrayAlmostEqual(output, expected, 3)
        # test just one input point
        output2 = lattice.get_all_distances(fcoords[0], fcoords)
        self.assertArrayAlmostEqual(output2, [expected[0]], 2)
        # test distance when initial points are not in unit cell
        f1 = [0, 0, 17]
        f2 = [0, 0, 10]
        self.assertEqual(lattice.get_all_distances(f1, f2)[0, 0], 0)

    def test_monoclinic(self):
        a, b, c, alpha, beta, gamma = self.monoclinic.parameters
        self.assertNotAlmostEqual(beta, 90)
        self.assertAlmostEqual(alpha, 90)
        self.assertAlmostEqual(gamma, 90)

    def test_is_hexagonal(self):
        self.assertFalse(self.cubic.is_hexagonal())
        self.assertFalse(self.tetragonal.is_hexagonal())
        self.assertFalse(self.orthorhombic.is_hexagonal())
        self.assertFalse(self.monoclinic.is_hexagonal())
        self.assertFalse(self.rhombohedral.is_hexagonal())
        self.assertTrue(self.hexagonal.is_hexagonal())

    def test_get_distance_and_image(self):
        dist, image = self.cubic.get_distance_and_image([0, 0, 0.1], [0, 0.0, 0.9])
        self.assertAlmostEqual(dist, 2)
        self.assertArrayAlmostEqual(image, [0, 0, -1])

    # def test_get_distance_and_image_strict(self):
    #     for count in range(10):
    #         lengths = [np.random.randint(1, 100) for i in range(3)]
    #         lattice = [np.random.rand(3) * lengths[i]
    #                    for i in range(3)]
    #         lattice = Lattice(np.array(lattice))
    #
    #         f1 = np.random.rand(3)
    #         f2 = np.random.rand(3)
    #
    #         scope = list(range(-3, 4))
    #         min_image_dist = (float("inf"), None)
    #         for image in itertools.product(scope, scope, scope):
    #             cart = lattice.get_cartesian_coords(f1 - (f2 + image))
    #             dist = np.dot(cart, cart) ** 0.5
    #             if dist < min_image_dist[0]:
    #                 min_image_dist = (dist, image)
    #
    #         pmg_result = lattice.get_distance_and_image(f1, f2)
    #         self.assertGreaterEqual(min_image_dist[0] + 1e-7, pmg_result[0])
    #         if abs(min_image_dist[0] - pmg_result[0]) < 1e-12:
    #             self.assertArrayAlmostEqual(min_image_dist[1], pmg_result[1])

    def test_lll_basis(self):
        a = np.array([1.0, 0.1, 0.0])
        b = np.array([0.0, 2.0, 0.0])
        c = np.array([0.0, 0.0, 3.0])

        l1 = Lattice([a, b, c])
        l2 = Lattice([a + b, b + c, c])

        ccoords = np.array([[1, 1, 2], [2, 2, 1.5]])
        l1_fcoords = l1.get_fractional_coords(ccoords)
        l2_fcoords = l2.get_fractional_coords(ccoords)

        self.assertArrayAlmostEqual(l1.matrix, l2.lll_matrix)
        self.assertArrayAlmostEqual(np.dot(l2.lll_mapping, l2.matrix), l1.matrix)

        self.assertArrayAlmostEqual(np.dot(l2_fcoords, l2.matrix), np.dot(l1_fcoords, l1.matrix))

        lll_fcoords = l2.get_lll_frac_coords(l2_fcoords)

        self.assertArrayAlmostEqual(lll_fcoords, l1_fcoords)
        self.assertArrayAlmostEqual(l1.get_cartesian_coords(lll_fcoords), np.dot(lll_fcoords, l2.lll_matrix))

        self.assertArrayAlmostEqual(l2.get_frac_coords_from_lll(lll_fcoords), l2_fcoords)

    def test_get_miller_index_from_sites(self):
        # test on a cubic system
        m = Lattice.cubic(1)
        s1 = np.array([0.5, -1.5, 3])
        s2 = np.array([0.5, 3.0, -1.5])
        s3 = np.array([2.5, 1.5, -4.0])
        self.assertEqual(m.get_miller_index_from_coords([s1, s2, s3]), (2, 1, 1))

        # test on a hexagonal system
        m = Lattice([[2.319, -4.01662582, 0.0], [2.319, 4.01662582, 0.0], [0.0, 0.0, 7.252]])

        s1 = np.array([2.319, 1.33887527, 6.3455])
        s2 = np.array([1.1595, 0.66943764, 4.5325])
        s3 = np.array([1.1595, 0.66943764, 0.9065])
        hkl = m.get_miller_index_from_coords([s1, s2, s3])
        self.assertEqual(hkl, (2, -1, 0))

        # test for previous failing structure
        m = Lattice([10, 0, 0, 0, 10, 0, 0, 0, 10])
        sites = [[0.5, 0.8, 0.8], [0.5, 0.4, 0.2], [0.5, 0.3, 0.7]]

        hkl = m.get_miller_index_from_coords(sites, coords_are_cartesian=False)
        self.assertEqual(hkl, (1, 0, 0))

        # test for more than 3 sites
        sites = [[0.5, 0.8, 0.8], [0.5, 0.4, 0.2], [0.5, 0.3, 0.7], [0.5, 0.1, 0.2]]

        hkl = m.get_miller_index_from_coords(sites, coords_are_cartesian=False)
        self.assertEqual(hkl, (1, 0, 0))

    def test_points_in_spheres(self):
        points = [[0.0, 0.0, 0.0], [2.0, 2.0, 2.0]]
        lattice = Lattice.cubic(3)
        center_points = [[1.5, 1.5, 1.5]]
        nns = get_points_in_spheres(
            all_coords=np.array(points),
            center_coords=np.array(center_points),
            r=3,
            pbc=np.array([0, 0, 0], dtype=int),
            lattice=lattice,
            numerical_tol=1e-8,
        )
        self.assertEqual(len(nns[0]), 2)  # two neighbors

        nns = get_points_in_spheres(
            all_coords=np.array(points),
            center_coords=np.array(center_points),
            r=3,
            pbc=[1, 1, 1],
            lattice=lattice,
            numerical_tol=1e-8,
            return_fcoords=True,
        )
        self.assertEqual(len(nns[0]), 12)

        nns = get_points_in_spheres(
            all_coords=np.array(points),
            center_coords=np.array(center_points),
            r=3,
            pbc=np.array([True, False, False], dtype=int),
            lattice=lattice,
        )
        self.assertEqual(len(nns[0]), 4)


if __name__ == "__main__":
    import unittest

    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


from pymatgen.analysis.prototypes import AflowPrototypeMatcher
from pymatgen.util.testing import PymatgenTest


class AflowPrototypeMatcherTest(PymatgenTest):
    def test_prototype_matching(self):
        af = AflowPrototypeMatcher()

        struct = self.get_structure("Sn")
        prototype = af.get_prototypes(struct)[0]

        self.assertDictEqual(
            prototype["tags"],
            {
                "aflow": "A_cF8_227_a",
                "mineral": "diamond",
                "pearson": "cF8",
                "strukturbericht": "A4",
            },
        )

        struct = self.get_structure("CsCl")
        prototype = af.get_prototypes(struct)[0]

        self.assertDictEqual(
            prototype["tags"],
            {
                "aflow": "AB_cP2_221_b_a",
                "mineral": "",
                "pearson": "cP2",
                "strukturbericht": "B2",
            },
        )

        struct = self.get_structure("Li2O")
        prototype = af.get_prototypes(struct)[0]

        self.assertDictEqual(
            prototype["tags"],
            {
                "aflow": "AB2_cF12_225_a_c",
                "mineral": "Fluorite",
                "pearson": "cF12",
                "strukturbericht": "C1",
            },
        )



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

import datetime
import unittest

import numpy as np

from pymatgen.core.structure import Molecule, Structure
from pymatgen.util.provenance import Author, HistoryNode, StructureNL

"""
Unit tests for StructureNL (SNL) format
"""

__author__ = "Anubhav Jain"
__credits__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Anubhav Jain"
__email__ = "ajain@lbl.gov"
__date__ = "2/14/13"


class StructureNLCase(unittest.TestCase):
    def setUp(self):
        # set up a Structure
        self.s = Structure(np.eye(3, 3) * 3, ["Fe"], [[0, 0, 0]])
        self.s2 = Structure(np.eye(3, 3) * 3, ["Al"], [[0, 0, 0]])
        self.mol = Molecule(["He"], [[0, 0, 0]])
        # set up BibTeX strings
        self.matproj = (
            "@misc{MaterialsProject,\ntitle = {{Materials " "Project}},\nurl = {http://www.materialsproject.org}\n}"
        )
        self.pmg = (
            "@article{Ong2013,\n author = {Ong, "
            "Shyue Ping and Richards, William Davidson and Jain, "
            "Anubhav and Hautier, Geoffroy and Kocher, "
            "Michael and Cholia, Shreyas and Gunter, Dan and Chevrier,"
            " Vincent L. and Persson, Kristin A. and Ceder, Gerbrand},"
            "\n doi = {10.1016/j.commatsci.2012.10.028},"
            "\n issn = {09270256},\n journal = {Computational "
            "Materials Science},\n month = feb,\n pages = {314--319},"
            "\n publisher = {Elsevier B.V.},"
            "\n title = {{Python Materials Genomics (pymatgen): A "
            "robust, open-source python library for materials "
            "analysis}},\n url = {http://linkinghub.elsevier"
            ".com/retrieve/pii/S0927025612006295},\n volume = {68},"
            "\n year = {2013}\n}"
        )
        repeat = "REPEAT" * 10000
        self.superlong = "@misc{SuperLong,\ntitle = {{" + repeat + "}}}"
        self.unicode_title = "@misc{Unicode_Title,\ntitle = {{A \u73ab is a rose}}}"
        self.junk = "This is junk text, not a BibTeX reference"

        # set up remarks
        self.remark_fail = [
            "This is a really long remark that is clearly invalid and must fail, don't you agree? It would be silly "
            "to allow remarks that went on forever and ever."
        ]

        # set up some authors
        self.hulk = [{"name": "Hulk", "email": "hulk@avengers.com"}]
        self.america = "Captain America <captainamerica@avengers.com>"
        self.thor = [("Thor", "thor@avengers.com")]
        self.duo = "Iron Man <ironman@avengers.com>, " "Black Widow <blackwidow@avengers.com>"

        # set up HistoryNodes
        self.valid_node = HistoryNode("DB 1", "www.db1URLgoeshere.com", {"db1_id": 12424})
        self.valid_node2 = {
            "name": "DB 2",
            "url": "www.db2URLgoeshere.com",
            "description": {"db2_id": 12424},
        }
        self.invalid_node = {"name": "DB 3", "url": "http://www.db3isnotavalidnode.com"}

    def test_authors(self):
        a = StructureNL(self.s, self.hulk, references=self.pmg)
        self.assertEqual(a.authors[0].name, "Hulk")
        self.assertEqual(a.authors[0].email, "hulk@avengers.com")

        a = StructureNL(self.s, self.america, references=self.pmg)
        self.assertEqual(a.authors[0].name, "Captain America")
        self.assertEqual(a.authors[0].email, "captainamerica@avengers.com")

        a = StructureNL(self.s, self.thor, references=self.pmg)
        self.assertEqual(a.authors[0].name, "Thor")
        self.assertEqual(a.authors[0].email, "thor@avengers.com")

        a = StructureNL(self.s, self.duo, references=self.pmg)
        self.assertEqual(a.authors[0].name, "Iron Man")
        self.assertEqual(a.authors[0].email, "ironman@avengers.com")
        self.assertEqual(a.authors[1].name, "Black Widow")
        self.assertEqual(a.authors[1].email, "blackwidow@avengers.com")
        StructureNL(self.s, self.hulk, references=self.pmg)

    def test_references(self):
        # An empty string should be OK
        StructureNL(self.s, self.hulk, references="")

        # An empty list should not work
        self.assertRaises(ValueError, StructureNL, self.s, self.hulk, references=[])

        # junk reference should not work
        self.assertRaises(ValueError, StructureNL, self.s, self.hulk, references=self.junk)

        # good references should be ok
        StructureNL(self.s, self.hulk, references=self.pmg)

        # unicode references should work
        StructureNL(self.s, self.hulk, references=self.unicode_title)

        # multi-line references should be OK
        StructureNL(self.s, self.hulk, references="\n".join([self.matproj, self.pmg]))

        # super long references are bad
        self.assertRaises(ValueError, StructureNL, self.s, self.hulk, references=self.superlong)

    def test_historynodes(self):
        a = StructureNL(self.s, self.hulk, history=[self.valid_node])
        self.assertEqual(a.history[0].name, "DB 1")
        self.assertEqual(a.history[0].url, "www.db1URLgoeshere.com")
        self.assertEqual(a.history[0].description, {"db1_id": 12424})

        a = StructureNL(self.s, self.hulk, history=[self.valid_node, self.valid_node2])
        self.assertEqual(a.history[1].name, "DB 2")
        self.assertEqual(a.history[1].url, "www.db2URLgoeshere.com")
        self.assertEqual(a.history[1].description, {"db2_id": 12424})

        # invalid nodes should not work
        self.assertRaises(Exception, StructureNL, self.s, self.hulk, history=[self.invalid_node])

        # too many nodes should not work
        self.assertRaises(ValueError, StructureNL, self.s, self.hulk, history=[self.valid_node] * 1000)

    def test_data(self):
        # Structure data is OK due to PMGEncoder/Decoder
        a = StructureNL(self.s, self.hulk, data={"_structure": self.s2})
        self.assertEqual(a.data["_structure"], self.s2, "Data storage is broken")
        self.assertRaises(ValueError, StructureNL, self.s, self.hulk, data={"bad_key": 1})

    def test_remarks(self):
        a = StructureNL(self.s, self.hulk, remarks="string format")
        self.assertEqual(a.remarks[0], "string format")
        self.assertRaises(ValueError, StructureNL, self.s, self.hulk, remarks=self.remark_fail)

    def test_eq(self):
        # test basic Equal()
        created_at = datetime.datetime.now()
        a = StructureNL(
            self.s,
            self.hulk,
            ["test_project"],
            self.pmg,
            ["remark1"],
            {"_my_data": self.s2},
            [self.valid_node, self.valid_node2],
            created_at,
        )
        b = StructureNL(
            self.s,
            self.hulk,
            ["test_project"],
            self.pmg,
            ["remark1"],
            {"_my_data": self.s2},
            [self.valid_node, self.valid_node2],
            created_at,
        )
        self.assertEqual(a, b, "__eq__() method is broken! false negative")

        # change the created at date, now they are no longer equal
        created_at = datetime.datetime.now() + datetime.timedelta(days=-1)
        c = StructureNL(
            self.s,
            self.hulk,
            ["test_project"],
            self.pmg,
            ["remark1"],
            {"_my_data": self.s2},
            [self.valid_node, self.valid_node2],
            created_at,
        )
        self.assertNotEqual(a, c, "__eq__() method is broken! false positive")

        # or try a different structure, those should not be equal
        d = StructureNL(
            self.s2,
            self.hulk,
            ["test_project"],
            self.pmg,
            ["remark1"],
            {"_my_data": self.s2},
            [self.valid_node, self.valid_node2],
            created_at,
        )
        self.assertNotEqual(a, d, "__eq__() method is broken! false positive")

    def test_to_from_dict(self):
        # no complicated objects in the 'data' or 'nodes' field
        a = StructureNL(
            self.s,
            self.hulk,
            ["test_project"],
            self.pmg,
            ["remark1"],
            {"_my_data": "string"},
            [self.valid_node, self.valid_node2],
        )
        b = StructureNL.from_dict(a.as_dict())
        self.assertEqual(a, b)
        # complicated objects in the 'data' and 'nodes' field
        complicated_node = {
            "name": "complicated node",
            "url": "www.complicatednodegoeshere.com",
            "description": {"structure": self.s2},
        }
        a = StructureNL(
            self.s,
            self.hulk,
            ["test_project"],
            self.pmg,
            ["remark1"],
            {"_my_data": {"structure": self.s2}},
            [complicated_node, self.valid_node],
        )
        b = StructureNL.from_dict(a.as_dict())
        self.assertEqual(
            a,
            b,
            "to/from dict is broken when object embedding is " "used! Apparently MontyEncoding is broken...",
        )

        # Test molecule
        molnl = StructureNL(self.mol, self.hulk, references=self.pmg)
        b = StructureNL.from_dict(molnl.as_dict())
        self.assertEqual(molnl, b)

    def test_from_structures(self):
        s1 = Structure([[5, 0, 0], [0, 5, 0], [0, 0, 5]], ["Fe"], [[0, 0, 0]])
        s2 = Structure([[5, 0, 0], [0, 5, 0], [0, 0, 5]], ["Mn"], [[0, 0, 0]])
        remarks = ["unittest"]
        authors = "Test User <test@materialsproject.com>"
        snl_list = StructureNL.from_structures([s1, s2], authors, remarks=remarks)

        self.assertEqual(len(snl_list), 2)
        snl1 = snl_list[0]
        snl2 = snl_list[1]
        self.assertEqual(snl1.remarks, remarks)
        self.assertEqual(snl2.remarks, remarks)
        self.assertEqual(snl1.authors, [Author.parse_author(authors)])
        self.assertEqual(snl2.authors, [Author.parse_author(authors)])


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


"""
Created on Jan 24, 2012
"""


__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__date__ = "Jan 24, 2012"

import os
import unittest

from pymatgen.core.structure import Structure
from pymatgen.io.cssr import Cssr
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.util.testing import PymatgenTest


class CssrTest(unittest.TestCase):
    def setUp(self):

        filepath = os.path.join(PymatgenTest.TEST_FILES_DIR, "POSCAR")
        p = Poscar.from_file(filepath)
        self.cssr = Cssr(p.structure)

    def test_str(self):
        expected_string = """10.4118 6.0672 4.7595
90.00 90.00 90.00 SPGR =  1 P 1    OPT = 1
24 0
0 Fe4 P4 O16
1 Fe 0.2187 0.7500 0.4749
2 Fe 0.2813 0.2500 0.9749
3 Fe 0.7187 0.7500 0.0251
4 Fe 0.7813 0.2500 0.5251
5 P 0.0946 0.2500 0.4182
6 P 0.4054 0.7500 0.9182
7 P 0.5946 0.2500 0.0818
8 P 0.9054 0.7500 0.5818
9 O 0.0434 0.7500 0.7071
10 O 0.0966 0.2500 0.7413
11 O 0.1657 0.0461 0.2854
12 O 0.1657 0.4539 0.2854
13 O 0.3343 0.5461 0.7854
14 O 0.3343 0.9539 0.7854
15 O 0.4034 0.7500 0.2413
16 O 0.4566 0.2500 0.2071
17 O 0.5434 0.7500 0.7929
18 O 0.5966 0.2500 0.7587
19 O 0.6657 0.0461 0.2146
20 O 0.6657 0.4539 0.2146
21 O 0.8343 0.5461 0.7146
22 O 0.8343 0.9539 0.7146
23 O 0.9034 0.7500 0.2587
24 O 0.9566 0.2500 0.2929"""
        self.assertEqual(str(self.cssr), expected_string)

    def test_from_file(self):
        filename = os.path.join(PymatgenTest.TEST_FILES_DIR, "Si.cssr")
        cssr = Cssr.from_file(filename)
        self.assertIsInstance(cssr.structure, Structure)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


import math
import os
import pickle
import unittest
import warnings
from copy import deepcopy

import numpy as np

from pymatgen.core.periodic_table import DummySpecies, Element, Species, get_el_sp
from pymatgen.util.testing import PymatgenTest


class ElementTestCase(PymatgenTest):
    def test_init(self):
        self.assertEqual("Fe", Element("Fe").symbol, "Fe test failed")

        fictional_symbols = ["D", "T", "Zebra"]

        for sym in fictional_symbols:
            self.assertRaises(ValueError, Element, sym)

        # Test caching
        self.assertEqual(id(Element("Fe")), id(Element("Fe")))

    def test_is_metal(self):
        for metal in ["Fe", "Eu", "Li", "Ca", "In"]:
            self.assertTrue(Element(metal).is_metal)
        for non_metal in ["Ge", "Si", "O", "He"]:
            self.assertFalse(Element(non_metal).is_metal)

    def test_nan_X(self):
        self.assertTrue(math.isnan(Element.He.X))
        els = sorted([Element.He, Element.H, Element.F])
        self.assertEqual(els, [Element.H, Element.F, Element.He])

    def test_dict(self):
        fe = Element.Fe
        d = fe.as_dict()
        self.assertEqual(fe, Element.from_dict(d))

    def test_block(self):
        testsets = {
            "O": "p",
            "Fe": "d",
            "Li": "s",
            "U": "f",
            "Er": "f",
            "Lu": "d",
            "Lr": "d",
        }
        for k, v in testsets.items():
            self.assertEqual(Element(k).block, v)

    def test_full_electronic_structure(self):
        testsets = {
            "O": [(1, "s", 2), (2, "s", 2), (2, "p", 4)],
            "Fe": [
                (1, "s", 2),
                (2, "s", 2),
                (2, "p", 6),
                (3, "s", 2),
                (3, "p", 6),
                (3, "d", 6),
                (4, "s", 2),
            ],
            "Li": [(1, "s", 2), (2, "s", 1)],
            "U": [
                (1, "s", 2),
                (2, "s", 2),
                (2, "p", 6),
                (3, "s", 2),
                (3, "p", 6),
                (3, "d", 10),
                (4, "s", 2),
                (4, "p", 6),
                (4, "d", 10),
                (5, "s", 2),
                (5, "p", 6),
                (4, "f", 14),
                (5, "d", 10),
                (6, "s", 2),
                (6, "p", 6),
                (5, "f", 3),
                (6, "d", 1),
                (7, "s", 2),
            ],
        }
        for k, v in testsets.items():
            self.assertEqual(Element(k).full_electronic_structure, v)

        self.assertEqual(Element.Ac.electronic_structure, "[Rn].6d1.7s2")

    def test_valence(self):
        testsets = {"O": (1, 4), "Fe": (2, 6), "Li": (0, 1), "Be": (0, 2)}
        for k, v in testsets.items():
            self.assertEqual(Element(k).valence, v)

        with self.assertRaises(ValueError):
            Element("U").valence

        valence = Element("He").valence
        self.assertTrue(np.isnan(valence[0]))
        self.assertEqual(valence[1], 0)

    def test_term_symbols(self):
        testsets = {
            "Li": [["2S0.5"]],  # s1
            "C": [["1D2.0"], ["3P0.0", "3P1.0", "3P2.0"], ["1S0.0"]],  # p2
            "Ti": [
                ["1G4.0"],
                ["3F2.0", "3F3.0", "3F4.0"],
                ["1D2.0"],
                ["3P0.0", "3P1.0", "3P2.0"],
                ["1S0.0"],
            ],  # d2
            "Pr": [
                ["2L7.5", "2L8.5"],
                ["2K6.5", "2K7.5"],
                ["4I4.5", "4I5.5", "4I6.5", "4I7.5"],
                ["2I5.5", "2I6.5"],
                ["2H4.5", "2H5.5"],
                ["2H4.5", "2H5.5"],
                ["4G2.5", "4G3.5", "4G4.5", "4G5.5"],
                ["2G3.5", "2G4.5"],
                ["2G3.5", "2G4.5"],
                ["4F1.5", "4F2.5", "4F3.5", "4F4.5"],
                ["2F2.5", "2F3.5"],
                ["2F2.5", "2F3.5"],
                ["4D0.5", "4D1.5", "4D2.5", "4D3.5"],
                ["2D1.5", "2D2.5"],
                ["2D1.5", "2D2.5"],
                ["2P0.5", "2P1.5"],
                ["4S1.5"],
            ],  # f3
        }
        for k, v in testsets.items():
            self.assertEqual(Element(k).term_symbols, v)

    def test_ground_state_term_symbol(self):
        testsets = {
            "Li": "2S0.5",  # s1
            "C": "3P0.0",  # p2
            "O": "3P2.0",  # p4
            "Ti": "3F2.0",  # d2
            "Pr": "4I4.5",
        }  # f3
        for k, v in testsets.items():
            self.assertEqual(Element(k).ground_state_term_symbol, v)

    def test_attributes(self):
        is_true = {
            ("Xe", "Kr"): "is_noble_gas",
            ("Fe", "Ni"): "is_transition_metal",
            ("Li", "Cs"): "is_alkali",
            ("Ca", "Mg"): "is_alkaline",
            ("F", "Br", "I"): "is_halogen",
            ("La",): "is_lanthanoid",
            ("U", "Pu"): "is_actinoid",
            ("Si", "Ge"): "is_metalloid",
            ("O", "Te"): "is_chalcogen",
        }

        for k, v in is_true.items():
            for sym in k:
                self.assertTrue(getattr(Element(sym), v), sym + " is false")

        keys = [
            "mendeleev_no",
            "atomic_mass",
            "electronic_structure",
            "atomic_radius",
            "min_oxidation_state",
            "max_oxidation_state",
            "electrical_resistivity",
            "velocity_of_sound",
            "reflectivity",
            "refractive_index",
            "poissons_ratio",
            "molar_volume",
            "thermal_conductivity",
            "melting_point",
            "boiling_point",
            "liquid_range",
            "critical_temperature",
            "superconduction_temperature",
            "bulk_modulus",
            "youngs_modulus",
            "brinell_hardness",
            "rigidity_modulus",
            "mineral_hardness",
            "vickers_hardness",
            "density_of_solid",
            "atomic_orbitals",
            "coefficient_of_linear_thermal_expansion",
            "oxidation_states",
            "common_oxidation_states",
            "average_ionic_radius",
            "average_cationic_radius",
            "average_anionic_radius",
            "ionic_radii",
            "long_name",
            "metallic_radius",
            "iupac_ordering",
            "ground_level",
            "ionization_energies",
        ]

        # Test all elements up to Uranium
        for i in range(1, 104):
            el = Element.from_Z(i)
            d = el.data
            for k in keys:
                k_str = k.capitalize().replace("_", " ")
                if k_str in d and (not str(d[k_str]).startswith("no data")):
                    self.assertIsNotNone(getattr(el, k))
                elif k == "long_name":
                    self.assertEqual(getattr(el, "long_name"), d["Name"])
                elif k == "iupac_ordering":
                    self.assertTrue("IUPAC ordering" in d)
                    self.assertIsNotNone(getattr(el, k))
            el = Element.from_Z(i)
            if len(el.oxidation_states) > 0:
                self.assertEqual(max(el.oxidation_states), el.max_oxidation_state)
                self.assertEqual(min(el.oxidation_states), el.min_oxidation_state)

            if el.symbol not in ["He", "Ne", "Ar"]:
                self.assertTrue(el.X > 0, "No electroneg for %s" % el)

        self.assertRaises(ValueError, Element.from_Z, 1000)

    def test_ie_ea(self):
        self.assertAlmostEqual(Element.Fe.ionization_energies[2], 30.651)
        self.assertEqual(Element.Fe.ionization_energy, Element.Fe.ionization_energies[0])
        self.assertAlmostEqual(Element.Br.electron_affinity, 3.3635883)

    def test_oxidation_states(self):
        el = Element.Fe
        self.assertEqual(el.oxidation_states, (-2, -1, 1, 2, 3, 4, 5, 6))
        self.assertEqual(el.common_oxidation_states, (2, 3))
        self.assertEqual(el.icsd_oxidation_states, (2, 3))

    def test_deepcopy(self):
        el1 = Element.Fe
        el2 = Element.Na
        ellist = [el1, el2]
        self.assertEqual(ellist, deepcopy(ellist), "Deepcopy operation doesn't produce exact copy")

    def test_radii(self):
        el = Element.Pd
        self.assertEqual(el.atomic_radius, 1.40)
        self.assertEqual(el.atomic_radius_calculated, 1.69)
        self.assertEqual(el.van_der_waals_radius, 1.63)

    def test_data(self):
        self.assertEqual(Element.Pd.data["Atomic radius"], 1.4)
        al = Element.Al
        val = al.thermal_conductivity
        self.assertEqual(val, 235)
        self.assertEqual(str(val.unit), "W K^-1 m^-1")
        val = al.electrical_resistivity
        self.assertEqual(val, 2.7e-08)
        self.assertEqual(str(val.unit), "m ohm")

    def test_sort(self):
        els = [Element.Se, Element.C]
        self.assertEqual(sorted(els), [Element.C, Element.Se])

    def test_pickle(self):
        el1 = Element.Fe
        o = pickle.dumps(el1)
        self.assertEqual(el1, pickle.loads(o))

        # Test all elements up to Uranium
        for i in range(1, 93):
            self.serialize_with_pickle(Element.from_Z(i), test_eq=True)

    def test_print_periodic_table(self):
        Element.print_periodic_table()

    def test_is(self):
        self.assertTrue(Element("Bi").is_post_transition_metal, True)


class SpecieTestCase(PymatgenTest):
    def setUp(self):
        self.specie1 = Species.from_string("Fe2+")
        self.specie2 = Species("Fe", 3)
        self.specie3 = Species("Fe", 2)
        self.specie4 = Species("Fe", 2, {"spin": 5})

    def test_init(self):
        self.assertRaises(ValueError, Species, "Fe", 2, {"magmom": 5})

    def test_ionic_radius(self):
        self.assertEqual(self.specie2.ionic_radius, 78.5 / 100)
        self.assertEqual(self.specie3.ionic_radius, 92 / 100)
        self.assertAlmostEqual(Species("Mn", 4).ionic_radius, 0.67)

    def test_eq(self):
        self.assertEqual(
            self.specie1,
            self.specie3,
            "Static and actual constructor gives unequal result!",
        )
        self.assertNotEqual(self.specie1, self.specie2, "Fe2+ should not be equal to Fe3+")
        self.assertNotEqual(self.specie4, self.specie3)
        self.assertFalse(self.specie1 == Element("Fe"))
        self.assertFalse(Element("Fe") == self.specie1)

    def test_cmp(self):
        self.assertLess(self.specie1, self.specie2, "Fe2+ should be < Fe3+")
        self.assertLess(Species("C", 1), Species("Se", 1))

    def test_attr(self):
        self.assertEqual(self.specie1.Z, 26, "Z attribute for Fe2+ should be = Element Fe.")
        self.assertEqual(self.specie4.spin, 5)

    def test_deepcopy(self):
        el1 = Species("Fe", 4)
        el2 = Species("Na", 1)
        ellist = [el1, el2]
        self.assertEqual(ellist, deepcopy(ellist), "Deepcopy operation doesn't produce exact copy.")

    def test_pickle(self):
        self.assertEqual(self.specie1, pickle.loads(pickle.dumps(self.specie1)))
        for i in range(1, 5):
            self.serialize_with_pickle(getattr(self, "specie%d" % i), test_eq=True)
        cs = Species("Cs", 1)
        cl = Species("Cl", 1)

        with open("cscl.pickle", "wb") as f:
            pickle.dump((cs, cl), f)

        with open("cscl.pickle", "rb") as f:
            d = pickle.load(f)
            self.assertEqual(d, (cs, cl))
        os.remove("cscl.pickle")

    def test_get_crystal_field_spin(self):
        self.assertEqual(Species("Fe", 2).get_crystal_field_spin(), 4)
        self.assertEqual(Species("Fe", 3).get_crystal_field_spin(), 5)
        self.assertEqual(Species("Fe", 4).get_crystal_field_spin(), 4)
        self.assertEqual(Species("Co", 3).get_crystal_field_spin(spin_config="low"), 0)
        self.assertEqual(Species("Co", 4).get_crystal_field_spin(spin_config="low"), 1)
        self.assertEqual(Species("Ni", 3).get_crystal_field_spin(spin_config="low"), 1)
        self.assertEqual(Species("Ni", 4).get_crystal_field_spin(spin_config="low"), 0)

        self.assertRaises(AttributeError, Species("Li", 1).get_crystal_field_spin)
        self.assertRaises(AttributeError, Species("Ge", 4).get_crystal_field_spin)
        self.assertRaises(AttributeError, Species("H", 1).get_crystal_field_spin)
        self.assertRaises(AttributeError, Species("Fe", 10).get_crystal_field_spin)
        self.assertRaises(ValueError, Species("Fe", 2).get_crystal_field_spin, "hex")

        s = Species("Co", 3).get_crystal_field_spin("tet", spin_config="low")
        self.assertEqual(s, 2)

    def test_get_nmr_mom(self):
        self.assertEqual(Species("H").get_nmr_quadrupole_moment(), 2.860)
        self.assertEqual(Species("Li").get_nmr_quadrupole_moment(), -0.808)
        self.assertEqual(Species("Li").get_nmr_quadrupole_moment("Li-7"), -40.1)
        self.assertEqual(Species("Si").get_nmr_quadrupole_moment(), 0.0)
        self.assertRaises(ValueError, Species("Li").get_nmr_quadrupole_moment, "Li-109")

    def test_get_shannon_radius(self):
        self.assertEqual(Species("Li", 1).get_shannon_radius("IV"), 0.59)
        mn2 = Species("Mn", 2)
        self.assertEqual(mn2.get_shannon_radius("IV", "High Spin"), 0.66)
        self.assertEqual(mn2.get_shannon_radius("V", "High Spin"), 0.75)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # Trigger a warning.
            r = mn2.get_shannon_radius("V")
            # Verify some things
            self.assertEqual(len(w), 1)
            self.assertIs(w[-1].category, UserWarning)
            self.assertEqual(r, 0.75)

        self.assertEqual(mn2.get_shannon_radius("VI", "Low Spin"), 0.67)
        self.assertEqual(mn2.get_shannon_radius("VI", "High Spin"), 0.83)
        self.assertEqual(mn2.get_shannon_radius("VII", "High Spin"), 0.9)
        self.assertEqual(mn2.get_shannon_radius("VIII"), 0.96)

    def test_sort(self):
        els = map(get_el_sp, ["N3-", "Si4+", "Si3+"])
        self.assertEqual(sorted(els), [Species("Si", 3), Species("Si", 4), Species("N", -3)])

    def test_to_from_string(self):
        fe3 = Species("Fe", 3, {"spin": 5})
        self.assertEqual(str(fe3), "Fe3+,spin=5")
        fe = Species.from_string("Fe3+,spin=5")
        self.assertEqual(fe.spin, 5)
        mo0 = Species("Mo", 0, {"spin": 5})
        self.assertEqual(str(mo0), "Mo0+,spin=5")
        mo = Species.from_string("Mo0+,spin=4")
        self.assertEqual(mo.spin, 4)

        # Shyue Ping: I don't understand the need for a None for oxidation state. That to me is basically an element.
        # Why make the thing so complicated for a use case that I have never seen???
        # fe_no_ox = Species("Fe", oxidation_state=None, properties={"spin": 5})
        # fe_no_ox_from_str = Species.from_string("Fe,spin=5")
        # self.assertEqual(fe_no_ox, fe_no_ox_from_str)

    def test_no_oxidation_state(self):
        mo0 = Species("Mo", None, {"spin": 5})
        self.assertEqual(str(mo0), "Mo,spin=5")

    def test_stringify(self):
        self.assertEqual(self.specie2.to_latex_string(), "Fe$^{3+}$")
        self.assertEqual(self.specie2.to_unicode_string(), "Fe³⁺")
        self.assertEqual(Species("S", -2).to_latex_string(), "S$^{2-}$")
        self.assertEqual(Species("S", -2).to_unicode_string(), "S²⁻")


class DummySpecieTestCase(unittest.TestCase):
    def test_init(self):
        self.specie1 = DummySpecies("X")
        self.assertRaises(ValueError, DummySpecies, "Xe")
        self.assertRaises(ValueError, DummySpecies, "Xec")
        self.assertRaises(ValueError, DummySpecies, "Vac")
        self.specie2 = DummySpecies("X", 2, {"spin": 3})
        self.assertEqual(self.specie2.spin, 3)

    def test_eq(self):
        self.assertFalse(DummySpecies("Xg") == DummySpecies("Xh"))
        self.assertFalse(DummySpecies("Xg") == DummySpecies("Xg", 3))
        self.assertTrue(DummySpecies("Xg", 3) == DummySpecies("Xg", 3))

    def test_from_string(self):
        sp = DummySpecies.from_string("X")
        self.assertEqual(sp.oxi_state, 0)
        sp = DummySpecies.from_string("X2+")
        self.assertEqual(sp.oxi_state, 2)
        self.assertEqual(sp.to_latex_string(), "X$^{2+}$")
        sp = DummySpecies.from_string("X2+spin=5")
        self.assertEqual(sp.oxi_state, 2)
        self.assertEqual(sp.spin, 5)
        self.assertEqual(sp.to_latex_string(), "X$^{2+}$")
        self.assertEqual(sp.to_html_string(), "X<sup>2+</sup>")
        self.assertEqual(sp.to_unicode_string(), "X²⁺")

    def test_pickle(self):
        el1 = DummySpecies("X", 3)
        o = pickle.dumps(el1)
        self.assertEqual(el1, pickle.loads(o))

    def test_sort(self):
        r = sorted([Element.Fe, DummySpecies("X")])
        self.assertEqual(r, [DummySpecies("X"), Element.Fe])
        self.assertTrue(DummySpecies("X", 3) < DummySpecies("X", 4))


class FuncTest(unittest.TestCase):
    def test_get_el_sp(self):
        self.assertEqual(get_el_sp("Fe2+"), Species("Fe", 2))
        self.assertEqual(get_el_sp("3"), Element.Li)
        self.assertEqual(get_el_sp("3.0"), Element.Li)
        self.assertEqual(get_el_sp("U"), Element.U)
        self.assertEqual(get_el_sp("X2+"), DummySpecies("X", 2))
        self.assertEqual(get_el_sp("Mn3+"), Species("Mn", 3))


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

__author__ = "Kyle Bystrom"
__copyright__ = "Copyright 2019, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Kyle Bystrom"
__email__ = "kylebystrom@gmail.com"
__date__ = "5/29/2019"

import random
import unittest

import numpy as np

from pymatgen.analysis.interface import Interface, InterfaceBuilder
from pymatgen.analysis.substrate_analyzer import ZSLGenerator
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.testing import PymatgenTest


class InterfaceTest(PymatgenTest):
    @classmethod
    def setUpClass(cls):
        si_struct = cls.get_structure("Si")
        sio2_struct = cls.get_structure("SiO2")

        sga = SpacegroupAnalyzer(si_struct)
        si_conventional = sga.get_conventional_standard_structure()
        sga = SpacegroupAnalyzer(sio2_struct)
        sio2_conventional = sga.get_conventional_standard_structure()
        cls.ib = InterfaceBuilder(si_conventional, sio2_conventional)
        cls.ib.generate_interfaces(substrate_millers=[[1, 0, 0]], film_layers=3, substrate_layers=3)

    def test_offset_vector(self):
        interface = self.ib.interfaces[0]
        init_lattice = interface.lattice.matrix.copy()
        self.assertArrayAlmostEqual(interface.offset_vector, np.array([0, 0, 2.5]))
        init_film = interface.film
        init_sub = interface.substrate
        tst = Structure.from_sites(interface.sites)

        interface.z_shift += 1
        self.assertArrayAlmostEqual(interface.offset_vector, np.array([0, 0, 3.5]))
        tdm, idm = tst.distance_matrix, interface.distance_matrix
        diff = tdm - idm
        assert (tdm <= idm + 1e-10).all()
        assert (tdm + 0.5 < idm).any()
        self.assertArrayAlmostEqual(init_film.distance_matrix, interface.film.distance_matrix)
        self.assertArrayAlmostEqual(init_sub.distance_matrix, interface.substrate.distance_matrix)

        interface.z_shift -= 1
        self.assertArrayAlmostEqual(interface.offset_vector, np.array([0, 0, 2.5]))
        idm = interface.distance_matrix
        assert (np.abs(tdm - idm) < 1e-10).all()

        interface.ab_shift += np.array([0.2, 0.2])
        self.assertArrayAlmostEqual(interface.ab_shift, np.array([0.2, 0.2]))
        idm = interface.distance_matrix
        assert (np.abs(tdm - idm) > 0.9).any()
        self.assertArrayAlmostEqual(init_lattice, interface.lattice.matrix)
        self.assertArrayAlmostEqual(init_film.distance_matrix, interface.film.distance_matrix)
        self.assertArrayAlmostEqual(init_sub.distance_matrix, interface.substrate.distance_matrix)

        interface.ab_shift -= np.array([0.2, 0.2])
        self.assertArrayAlmostEqual(interface.offset_vector, np.array([0, 0, 2.5]))
        idm = interface.distance_matrix
        assert (np.abs(tdm - idm) < 1e-10).all()

        self.assertArrayAlmostEqual(init_film.distance_matrix, interface.film.distance_matrix)
        self.assertArrayAlmostEqual(init_sub.distance_matrix, interface.substrate.distance_matrix)

    def test_labels(self):
        interface = self.ib.interfaces[0]
        film = interface.film
        substrate = interface.substrate
        film_sites = [
            site for i, site in enumerate(interface) if "film" in interface.site_properties["interface_label"][i]
        ]
        substrate_sites = [
            site for i, site in enumerate(interface) if "substrate" in interface.site_properties["interface_label"][i]
        ]
        assert film.sites == film_sites
        assert substrate.sites == substrate_sites
        assert len(film) == len(interface.modified_film_structure)
        assert len(substrate) == len(interface.modified_sub_structure)

    def test_vertical_spacing(self):
        interface = self.ib.interfaces[0]
        self.assertAlmostEqual(
            interface.z_shift,
            np.min(interface.film.cart_coords[:, 2]) - np.max(interface.substrate.cart_coords[:, 2]),
        )
        self.assertAlmostEqual(
            interface.lattice.c,
            interface.vacuum_thickness
            + interface.z_shift
            + np.max(interface.film.cart_coords[:, 2])
            - np.min(interface.film.cart_coords[:, 2])
            + np.max(interface.substrate.cart_coords[:, 2])
            - np.min(interface.substrate.cart_coords[:, 2]),
        )

    def test_inplane_spacing(self):
        delta = np.array([0, 1.5, 0])
        interface = self.ib.interfaces[0]
        old_coords = interface.film.frac_coords.copy()
        interface.offset_vector += delta
        self.assertArrayAlmostEqual(interface.offset_vector, [0, 1.5, 2.5])
        new_coords = interface.film.frac_coords.copy()
        for i in range(new_coords.shape[0]):
            self.assertAlmostEqual(
                interface.lattice.get_distance_and_image(old_coords[i], new_coords[i])[0],
                1.5,
            )
        interface.offset_vector -= delta
        self.assertArrayAlmostEqual(interface.offset_vector, [0, 0, 2.5])
        new_coords = interface.film.frac_coords.copy()
        for i in range(new_coords.shape[0]):
            self.assertAlmostEqual(
                interface.lattice.get_distance_and_image(old_coords[i], new_coords[i])[0],
                0.0,
            )

    def test_copy(self):
        interface = self.ib.interfaces[0]
        copy = interface.copy()
        for attr in [
            "lattice",
            "cart_coords",
            "sub_plane",
            "film_plane",
            "modified_film_structure",
            "modified_sub_structure",
            "strained_film_structure",
            "strained_sub_structure",
            "sub_init_cell",
            "film_init_cell",
            "site_properties",
            "offset_vector",
            "ab_shift",
            "z_shift",
            "vacuum_thickness",
        ]:
            if type(copy.__getattribute__(attr)) == np.ndarray:
                self.assertArrayAlmostEqual(copy.__getattribute__(attr), interface.__getattribute__(attr))
            else:
                assert copy.__getattribute__(attr) == interface.__getattribute__(attr)

    def test_serialization(self):
        interface = self.ib.interfaces[0]
        interface_dict = interface.as_dict()
        interface_from_dict = Interface.from_dict(interface_dict)
        for attr in [
            "lattice",
            "cart_coords",
            "sub_plane",
            "film_plane",
            "modified_film_structure",
            "modified_sub_structure",
            "strained_film_structure",
            "strained_sub_structure",
            "sub_init_cell",
            "film_init_cell",
            "site_properties",
            "offset_vector",
            "ab_shift",
            "z_shift",
            "vacuum_thickness",
        ]:
            if type(interface_from_dict.__getattribute__(attr)) == np.ndarray:
                self.assertArrayAlmostEqual(
                    interface_from_dict.__getattribute__(attr),
                    interface.__getattribute__(attr),
                )
            else:
                self.assertAlmostEqual(
                    interface_from_dict.__getattribute__(attr),
                    interface.__getattribute__(attr),
                )

        # Shift film and check equality
        interface = self.ib.interfaces[0].copy()
        interface.z_shift = 4
        interface.ab_shift = [0.5, 0.5]
        interface_dict = interface.as_dict()
        interface_from_dict = Interface.from_dict(interface_dict)
        for attr in [
            "lattice",
            "cart_coords",
            "sub_plane",
            "film_plane",
            "modified_film_structure",
            "modified_sub_structure",
            "strained_film_structure",
            "strained_sub_structure",
            "sub_init_cell",
            "film_init_cell",
            "site_properties",
            "offset_vector",
            "ab_shift",
            "z_shift",
            "vacuum_thickness",
        ]:
            if type(interface_from_dict.__getattribute__(attr)) == np.ndarray:
                self.assertArrayAlmostEqual(
                    interface_from_dict.__getattribute__(attr),
                    interface.__getattribute__(attr),
                )
            else:
                self.assertAlmostEqual(
                    interface_from_dict.__getattribute__(attr),
                    interface.__getattribute__(attr),
                )


class InterfaceBuilderTest(PymatgenTest):
    @classmethod
    def setUpClass(cls):
        si_struct = cls.get_structure("Si")
        sio2_struct = cls.get_structure("SiO2")
        sga = SpacegroupAnalyzer(si_struct)
        si_conventional = sga.get_conventional_standard_structure()
        sga = SpacegroupAnalyzer(sio2_struct)
        sio2_conventional = sga.get_conventional_standard_structure()
        cls.ibs = []
        Si_SiO2 = [si_conventional, sio2_conventional]
        random.shuffle(Si_SiO2)
        cls.ibs.append(cls.make_ib(*Si_SiO2, [1, 0, 0]))
        Si_SiO2 = [si_struct, sio2_struct]
        random.shuffle(Si_SiO2)
        cls.ibs.append(cls.make_ib(*Si_SiO2, [1, 0, 0]))
        cls.ibs.append(cls.make_ib(*Si_SiO2, [1, 1, 1]))

    @staticmethod
    def make_ib(substrate, film, miller):
        ib = InterfaceBuilder(substrate, film)
        ib.generate_interfaces(substrate_millers=[miller])
        return ib

    def test_film_and_substrate_sites(self):
        for ib in self.ibs:
            interface = ib.interfaces[0]
            assert len(interface.film) == len(interface.modified_film_structure)
            assert len(interface.substrate) == len(interface.modified_sub_structure)

    def test_lattice(self):
        zsl = ZSLGenerator()
        for ib in self.ibs:
            interface = ib.interfaces[0]
            assert zsl.is_same_vectors(
                interface.modified_sub_structure.lattice.matrix[:2],
                interface.modified_film_structure.lattice.matrix[:2],
            )

    def test_structure_preservation(self):
        for ib in self.ibs:
            for interface in ib.interfaces[:2]:
                # assumes test structures are SiO2 and Si
                substrate, film = interface.substrate, interface.film
                if substrate.ntypesp == 1:
                    si = substrate
                    sio2 = film
                else:
                    si = film
                    sio2 = substrate
                sidm = si.distance_matrix
                sidm = sidm[sidm > 0]
                sio2dm = sio2.distance_matrix
                sio2dm = sio2dm[sio2dm > 0]
                assert si.is_valid(tol=2.32)
                assert sio2.is_valid(tol=1.6)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from monty.serialization import loadfn

from pymatgen.util.testing import PymatgenTest


class IRDielectricTensorTest(PymatgenTest):
    def setUp(self):
        self.ir_spectra = loadfn(os.path.join(PymatgenTest.TEST_FILES_DIR, "ir_spectra_mp-991652_DDB.json"))

    def test_basic(self):
        self.ir_spectra.write_json("test.json")
        ir_spectra = loadfn("test.json")
        irdict = ir_spectra.as_dict()
        ir_spectra.from_dict(irdict)
        ir_spectra.plot(show=False)

    def tearDown(self):
        if os.path.isfile("test.json"):
            os.remove("test.json")


if __name__ == "__main__":
    unittest.main()
